"""
The RadarCollectionType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union, Optional, Dict, Tuple
import logging

import numpy

from sarpy.io.xml.base import Serializable, Arrayable, SerializableArray, \
    ParametersCollection, parse_float
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    FloatDescriptor, IntegerDescriptor, SerializableDescriptor, \
    SerializableArrayDescriptor, UnitVectorDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT, \
    SerializableCPArrayDescriptor, SerializableCPArray
from .blocks import XYZType, LatLonHAECornerRestrictionType, \
    POLARIZATION1_VALUES, POLARIZATION2_VALUES, DUAL_POLARIZATION_VALUES
from .utils import polstring_version_required

import sarpy.geometry.geocoords as geocoords


logger = logging.getLogger(__name__)


def get_band_name(freq: float) -> str:
    """
    Gets the band names associated with the given frequency (in Hz).

    Parameters
    ----------
    freq : float
        The frequency in Hz.

    Returns
    -------
    str
    """

    if freq is None:
        return 'UN'
    elif 3e6 <= freq < 3e7:
        return 'HF'
    elif 3e7 <= freq < 3e8:
        return 'VHF'
    elif 3e8 <= freq < 1e9:
        return 'UHF'
    elif 1e9 <= freq < 2e9:
        return 'L'
    elif 2e9 <= freq < 4e9:
        return 'S'
    elif 4e9 <= freq < 8e9:
        return 'C'
    elif 8e9 <= freq < 1.2e10:
        return 'X'
    elif 1.2e10 <= freq < 1.8e10:
        return 'KU'
    elif 1.8e10 <= freq < 2.7e10:
        return 'K'
    elif 2.7e10 <= freq < 4e10:
        return 'KA'
    elif 4e10 <= freq < 7.5e10:
        return 'V'
    elif 7.5e10 <= freq < 1.1e11:
        return 'W'
    elif 1.1e11 <= freq < 3e11:
        return 'MM'
    else:
        return 'UN'


class TxFrequencyType(Serializable, Arrayable):
    """
    The transmit frequency range.
    """

    _fields = ('Min', 'Max')
    _required = _fields
    _numeric_format = {'Min': FLOAT_FORMAT, 'Max': FLOAT_FORMAT}
    # descriptors
    Min = FloatDescriptor(
        'Min', _required, strict=DEFAULT_STRICT,
        docstring='The transmit minimum frequency in Hz.')  # type: float
    Max = FloatDescriptor(
        'Max', _required, strict=DEFAULT_STRICT,
        docstring='The transmit maximum frequency in Hz.')  # type: float

    def __init__(
            self,
            Min: float = None,
            Max: float = None,
            **kwargs):
        """

        Parameters
        ----------
        Min : float
        Max : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Min, self.Max = Min, Max
        super(TxFrequencyType, self).__init__(**kwargs)

    @property
    def center_frequency(self) -> Optional[float]:
        """
        None|float: The center frequency
        """

        if self.Min is None or self.Max is None:
            return None
        return 0.5*(self.Min + self.Max)

    def _apply_reference_frequency(self, reference_frequency: float):
        if self.Min is not None:
            self.Min += reference_frequency
        if self.Max is not None:
            self.Max += reference_frequency

    def _basic_validity_check(self) -> bool:
        condition = super(TxFrequencyType, self)._basic_validity_check()
        if self.Min is not None and self.Max is not None and self.Max < self.Min:
            self.log_validity_error(
                'Invalid frequency bounds Min ({}) > Max ({})'.format(self.Min, self.Max))
            condition = False
        return condition

    def get_band_abbreviation(self) -> str:
        """
        Gets the band abbreviation for the suggested name.

        Returns
        -------
        str
        """

        band_name = get_band_name(self.center_frequency)
        return band_name + '_'*(3 - len(band_name))

    def get_array(self, dtype=numpy.float64) -> numpy.ndarray:
        """
        Gets an array representation of the data.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            data type of the return

        Returns
        -------
        numpy.ndarray
            data array with appropriate entry order
        """

        return numpy.array([self.Min, self.Max], dtype=dtype)

    @classmethod
    def from_array(cls, array: Union[numpy.ndarray, list, tuple]):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Min, Max]

        Returns
        -------
        LatLonType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(len(array)))
            return cls(Min=array[0], Max=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class WaveformParametersType(Serializable):
    """
    Transmit and receive demodulation waveform parameters.
    """

    _fields = (
        'TxPulseLength', 'TxRFBandwidth', 'TxFreqStart', 'TxFMRate', 'RcvDemodType', 'RcvWindowLength',
        'ADCSampleRate', 'RcvIFBandwidth', 'RcvFreqStart', 'RcvFMRate', 'index')
    _required = ('index', )
    _set_as_attribute = ('index', )
    _numeric_format = {
        'TxPulseLength': FLOAT_FORMAT, 'TxRFBandwidth': '0.17E', 'TxFreqStart': '0.17E',
        'TxFMRate': '0.17E', 'RcvWindowLength': FLOAT_FORMAT, 'ADCSampleRate': '0.17E',
        'RcvIFBandwidth': '0.17E', 'RcvFreqStart': '0.17E', 'RcvFMRate': '0.17E'}

    # descriptors
    TxPulseLength = FloatDescriptor(
        'TxPulseLength', _required, strict=DEFAULT_STRICT,
        docstring='Transmit pulse length in seconds.')  # type: float
    TxRFBandwidth = FloatDescriptor(
        'TxRFBandwidth', _required, strict=DEFAULT_STRICT,
        docstring='Transmit RF bandwidth of the transmit pulse in Hz.')  # type: float
    TxFreqStart = FloatDescriptor(
        'TxFreqStart', _required, strict=DEFAULT_STRICT,
        docstring='Transmit Start frequency for Linear FM waveform in Hz, may be relative '
                  'to reference frequency.')  # type: float
    TxFMRate = FloatDescriptor(
        'TxFMRate', _required, strict=DEFAULT_STRICT,
        docstring='Transmit FM rate for Linear FM waveform in Hz/second.')  # type: float
    RcvWindowLength = FloatDescriptor(
        'RcvWindowLength', _required, strict=DEFAULT_STRICT,
        docstring='Receive window duration in seconds.')  # type: float
    ADCSampleRate = FloatDescriptor(
        'ADCSampleRate', _required, strict=DEFAULT_STRICT,
        docstring='Analog-to-Digital Converter sampling rate in samples/second.')  # type: float
    RcvIFBandwidth = FloatDescriptor(
        'RcvIFBandwidth', _required, strict=DEFAULT_STRICT,
        docstring='Receive IF bandwidth in Hz.')  # type: float
    RcvFreqStart = FloatDescriptor(
        'RcvFreqStart', _required, strict=DEFAULT_STRICT,
        docstring='Receive demodulation start frequency in Hz, may be relative to reference frequency.')  # type: float
    index = IntegerDescriptor(
        'index', _required, strict=False, docstring="The array index.")  # type: int

    def __init__(
            self,
            TxPulseLength: Optional[float] = None,
            TxRFBandwidth: Optional[float] = None,
            TxFreqStart: Optional[float] = None,
            TxFMRate: Optional[float] = None,
            RcvDemodType: Optional[str] = None,
            RcvWindowLength: Optional[float] = None,
            ADCSampleRate: Optional[float] = None,
            RcvIFBandwidth: Optional[float] = None,
            RcvFreqStart: Optional[float] = None,
            RcvFMRate: Optional[float] = None,
            index: int = None,
            **kwargs):
        """

        Parameters
        ----------
        TxPulseLength : float
        TxRFBandwidth : float
        TxFreqStart : float
        TxFMRate : float
        RcvDemodType : str
        RcvWindowLength : float
        ADCSampleRate : float
        RcvIFBandwidth : float
        RcvFreqStart : float
        RcvFMRate : float
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self._RcvFMRate = None
        self.TxPulseLength, self.TxRFBandwidth = TxPulseLength, TxRFBandwidth
        self.TxFreqStart, self.TxFMRate = TxFreqStart, TxFMRate
        self.RcvWindowLength = RcvWindowLength
        self.ADCSampleRate = ADCSampleRate
        self.RcvIFBandwidth = RcvIFBandwidth
        self.RcvFreqStart = RcvFreqStart
        # NB: self.RcvDemodType is read only.
        if RcvDemodType == 'CHIRP' and RcvFMRate is None:
            self.RcvFMRate = 0.0
        elif RcvDemodType == 'STRETCH' and RcvFMRate is not None:
            self.RcvFMRate = RcvFMRate
        else:
            self.RcvFMRate = None
        self.index = index
        super(WaveformParametersType, self).__init__(**kwargs)

    @property
    def RcvDemodType(self) -> Optional[str]:
        """
        str: READ ONLY. Receive demodulation used when Linear FM waveform is
        used on transmit. This value is derived form the value of `RcvFMRate`.

        * `None` - `RcvFMRate` is `None`.

        * `'CHIRP'` - `RcvFMRate=0`.

        * `'STRETCH'` - `RcvFMRate` is non-zero.
        """

        if self._RcvFMRate is None:
            return None
        elif self._RcvFMRate == 0:
            return 'CHIRP'
        else:
            return 'STRETCH'

    @property
    def RcvFMRate(self) -> Optional[float]:
        """
        float: Receive FM rate in Hz/sec. Also, determines the value of `RcvDemodType`. **Optional.**
        """
        return self._RcvFMRate

    @RcvFMRate.setter
    def RcvFMRate(self, value: Optional[float]):
        if value is None:
            self._RcvFMRate = None
        else:
            try:
                self._RcvFMRate = parse_float(value, 'RcvFMRate', self)
            except Exception as e:
                logger.error(
                    'Failed parsing value {} for field RCVFMRate of type "float",\n\t'
                    'with error {} - {}.\n\t'
                    'The value has been set to None.'.format(value, type(e), e))
                self._RcvFMRate = None

    def _basic_validity_check(self) -> bool:
        valid = super(WaveformParametersType, self)._basic_validity_check()
        return valid

    def derive(self):
        """
        Populate derived data in `WaveformParametersType`.

        Returns
        -------
        None
        """

        if self.TxPulseLength is not None and self.TxFMRate is not None and self.TxRFBandwidth is None:
            self.TxRFBandwidth = self.TxPulseLength*self.TxFMRate
        if self.TxPulseLength is not None and self.TxRFBandwidth is not None and self.TxFMRate is None:
            self.TxFMRate = self.TxRFBandwidth/self.TxPulseLength
        if self.TxFMRate is not None and self.TxRFBandwidth is not None and self.TxPulseLength is None:
            self.TxPulseLength = self.TxRFBandwidth/self.TxFMRate

    def _apply_reference_frequency(self, reference_frequency: float):
        if self.TxFreqStart is not None:
            self.TxFreqStart += reference_frequency
        if self.RcvFreqStart is not None:
            self.RcvFreqStart += reference_frequency


class TxStepType(Serializable):
    """
    Transmit sequence step details.
    """

    _fields = ('WFIndex', 'TxPolarization', 'index')
    _required = ('index', )
    _set_as_attribute = ('index', )
    # descriptors
    WFIndex = IntegerDescriptor(
        'WFIndex', _required, strict=DEFAULT_STRICT,
        docstring='The waveform number for this step.')  # type: int
    TxPolarization = StringEnumDescriptor(
        'TxPolarization', POLARIZATION1_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Transmit signal polarization for this step.')  # type: str
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT,
        docstring='The step index')  # type: int

    def __init__(
            self,
            WFIndex: Optional[int] = None,
            TxPolarization: Optional[str] = None,
            index: int = None,
            **kwargs):
        """

        Parameters
        ----------
        WFIndex : int
        TxPolarization : str
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.WFIndex = WFIndex
        self.TxPolarization = TxPolarization
        self.index = index
        super(TxStepType, self).__init__(**kwargs)


class ChanParametersType(Serializable):
    """
    Transmit receive sequence step details.
    """

    _fields = ('TxRcvPolarization', 'RcvAPCIndex', 'index')
    _required = ('TxRcvPolarization', 'index', )
    _set_as_attribute = ('index', )
    # descriptors
    TxRcvPolarization = StringEnumDescriptor(
        'TxRcvPolarization', DUAL_POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Combined Transmit and Receive signal polarization for the channel.')  # type: str
    RcvAPCIndex = IntegerDescriptor(
        'RcvAPCIndex', _required, strict=DEFAULT_STRICT,
        docstring='Index of the Receive Aperture Phase Center (Rcv APC). Only include if Receive APC position '
                  'polynomial(s) are included.')  # type: int
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT, docstring='The parameter index')  # type: int

    def __init__(
            self,
            TxRcvPolarization: Optional[str] = None,
            RcvAPCIndex: Optional[int] = None,
            index: int = None,
            **kwargs):
        """

        Parameters
        ----------
        TxRcvPolarization : str
        RcvAPCIndex : int
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxRcvPolarization = TxRcvPolarization
        self.RcvAPCIndex = RcvAPCIndex
        self.index = index
        super(ChanParametersType, self).__init__(**kwargs)

    def get_transmit_polarization(self) -> Optional[str]:
        if self.TxRcvPolarization is None:
            return None
        elif self.TxRcvPolarization in ['OTHER', 'UNKNOWN']:
            return 'OTHER'
        else:
            return self.TxRcvPolarization.split(':')[0]

    def version_required(self) -> Tuple[int, int, int]:
        """
        What SICD version is required?

        Returns
        -------
        tuple
        """

        return polstring_version_required(self.TxRcvPolarization)


class ReferencePointType(Serializable):
    """The reference point definition"""
    _fields = ('ECF', 'Line', 'Sample', 'name')
    _required = ('ECF', 'Line', 'Sample')
    _set_as_attribute = ('name', )
    _numeric_format = {'Line': FLOAT_FORMAT, 'Sample': FLOAT_FORMAT}
    # descriptors
    ECF = SerializableDescriptor(
        'ECF', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='The geographical coordinates for the reference point.')  # type: XYZType
    Line = FloatDescriptor(
        'Line', _required, strict=DEFAULT_STRICT,
        docstring='The reference point line index.')  # type: float
    Sample = FloatDescriptor(
        'Sample', _required, strict=DEFAULT_STRICT,
        docstring='The reference point sample index.')  # type: float
    name = StringDescriptor(
        'name', _required, strict=DEFAULT_STRICT,
        docstring='The reference point name.')  # type: str

    def __init__(
            self,
            ECF: Union[XYZType, numpy.ndarray, list, tuple] = None,
            Line: float = None,
            Sample: float = None,
            name: Optional[str] = None,
            **kwargs):
        """

        Parameters
        ----------
        ECF : XYZType|numpy.ndarray|list|tuple
        Line : float
        Sample : float
        name : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ECF = ECF
        self.Line = Line
        self.Sample = Sample
        self.name = name
        super(ReferencePointType, self).__init__(**kwargs)


class XDirectionType(Serializable):
    """The X direction of the collect"""
    _fields = ('UVectECF', 'LineSpacing', 'NumLines', 'FirstLine')
    _required = _fields
    _numeric_format = {'LineSpacing': FLOAT_FORMAT, }
    # descriptors
    UVectECF = UnitVectorDescriptor(
        'UVectECF', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='The unit vector in the X direction.')  # type: XYZType
    LineSpacing = FloatDescriptor(
        'LineSpacing', _required, strict=DEFAULT_STRICT,
        docstring='The collection line spacing in the X direction in meters.')  # type: float
    NumLines = IntegerDescriptor(
        'NumLines', _required, strict=DEFAULT_STRICT,
        docstring='The number of lines in the X direction.')  # type: int
    FirstLine = IntegerDescriptor(
        'FirstLine', _required, strict=DEFAULT_STRICT,
        docstring='The first line index.')  # type: int

    def __init__(
            self,
            UVectECF: Union[XYZType, numpy.ndarray, list, tuple] = None,
            LineSpacing: int = None,
            NumLines: int = None,
            FirstLine: int = None,
            **kwargs):
        """

        Parameters
        ----------
        UVectECF : XYZType|numpy.ndarray|list|tuple
        LineSpacing : float
        NumLines : int
        FirstLine : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.UVectECF = UVectECF
        self.LineSpacing = LineSpacing
        self.NumLines = NumLines
        self.FirstLine = FirstLine
        super(XDirectionType, self).__init__(**kwargs)


class YDirectionType(Serializable):
    """The Y direction of the collect"""
    _fields = ('UVectECF', 'SampleSpacing', 'NumSamples', 'FirstSample')
    _required = _fields
    _numeric_format = {'SampleSpacing': FLOAT_FORMAT, }
    # descriptors
    UVectECF = UnitVectorDescriptor(
        'UVectECF', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='The unit vector in the Y direction.')  # type: XYZType
    SampleSpacing = FloatDescriptor(
        'SampleSpacing', _required, strict=DEFAULT_STRICT,
        docstring='The collection sample spacing in the Y direction in meters.')  # type: float
    NumSamples = IntegerDescriptor(
        'NumSamples', _required, strict=DEFAULT_STRICT,
        docstring='The number of samples in the Y direction.')  # type: int
    FirstSample = IntegerDescriptor(
        'FirstSample', _required, strict=DEFAULT_STRICT,
        docstring='The first sample index.')  # type: int

    def __init__(
            self,
            UVectECF: Union[XYZType, numpy.ndarray, list, tuple] = None,
            SampleSpacing: float = None,
            NumSamples: int = None,
            FirstSample: int = None,
            **kwargs):
        """

        Parameters
        ----------
        UVectECF : XYZType|numpy.ndarray|list|tuple
        SampleSpacing : float
        NumSamples : int
        FirstSample : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.UVectECF = UVectECF
        self.SampleSpacing = SampleSpacing
        self.NumSamples = NumSamples
        self.FirstSample = FirstSample
        super(YDirectionType, self).__init__(**kwargs)


class SegmentArrayElement(Serializable):
    """The reference point definition"""
    _fields = ('StartLine', 'StartSample', 'EndLine', 'EndSample', 'Identifier', 'index')
    _required = _fields
    _set_as_attribute = ('index', )
    # descriptors
    StartLine = IntegerDescriptor(
        'StartLine', _required, strict=DEFAULT_STRICT,
        docstring='The starting line number of the segment.')  # type: int
    StartSample = IntegerDescriptor(
        'StartSample', _required, strict=DEFAULT_STRICT,
        docstring='The starting sample number of the segment.')  # type: int
    EndLine = IntegerDescriptor(
        'EndLine', _required, strict=DEFAULT_STRICT,
        docstring='The ending line number of the segment.')  # type: int
    EndSample = IntegerDescriptor(
        'EndSample', _required, strict=DEFAULT_STRICT,
        docstring='The ending sample number of the segment.')  # type: int
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='Identifier for the segment data boundary.')
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT,
        docstring='The array index.')  # type: int

    def __init__(
            self,
            StartLine: int = None,
            StartSample: int = None,
            EndLine: int = None,
            EndSample: int = None,
            Identifier: str = None,
            index: int = None,
            **kwargs):
        """

        Parameters
        ----------
        StartLine : int
        StartSample : int
        EndLine : int
        EndSample : int
        Identifier : str
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.StartLine, self.EndLine = StartLine, EndLine
        self.StartSample, self.EndSample = StartSample, EndSample
        self.Identifier = Identifier
        self.index = index
        super(SegmentArrayElement, self).__init__(**kwargs)


class ReferencePlaneType(Serializable):
    """
    The reference plane.
    """

    _fields = ('RefPt', 'XDir', 'YDir', 'SegmentList', 'Orientation')
    _required = ('RefPt', 'XDir', 'YDir')
    _collections_tags = {'SegmentList': {'array': True, 'child_tag': 'Segment'}}
    # other class variable
    _ORIENTATION_VALUES = ('UP', 'DOWN', 'LEFT', 'RIGHT', 'ARBITRARY')
    # descriptors
    RefPt = SerializableDescriptor(
        'RefPt', ReferencePointType, _required, strict=DEFAULT_STRICT,
        docstring='The reference point.')  # type: ReferencePointType
    XDir = SerializableDescriptor(
        'XDir', XDirectionType, _required, strict=DEFAULT_STRICT,
        docstring='The X direction collection plane parameters.')  # type: XDirectionType
    YDir = SerializableDescriptor(
        'YDir', YDirectionType, _required, strict=DEFAULT_STRICT,
        docstring='The Y direction collection plane parameters.')  # type: YDirectionType
    SegmentList = SerializableArrayDescriptor(
        'SegmentList', SegmentArrayElement, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The segment array.')  # type: Union[SerializableArray, List[SegmentArrayElement]]
    Orientation = StringEnumDescriptor(
        'Orientation', _ORIENTATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Describes the shadow intent of the display plane.')  # type: str

    def __init__(
            self,
            RefPt: ReferencePointType = None,
            XDir: XDirectionType = None,
            YDir: YDirectionType = None,
            SegmentList: Union[SerializableArray, List[SegmentArrayElement]] = None,
            Orientation: Optional[str] = None,
            **kwargs):
        """

        Parameters
        ----------
        RefPt : ReferencePointType
        XDir : XDirectionType
        YDir : YDirectionType
        SegmentList : SerializableArray|List[SegmentArrayElement]
        Orientation : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RefPt = RefPt
        self.XDir, self.YDir = XDir, YDir
        self.SegmentList = SegmentList
        self.Orientation = Orientation
        super(ReferencePlaneType, self).__init__(**kwargs)

    def get_ecf_corner_array(self) -> numpy.ndarray:
        """
        Use the XDir and YDir definitions to return the corner points in ECF coordinates as a `4x3` array.

        Returns
        -------
        numpy.ndarray
            The corner points of the collection area, with order following the AreaType order convention.
        """

        ecf_ref = self.RefPt.ECF.get_array()
        x_shift = self.XDir.UVectECF.get_array() * self.XDir.LineSpacing
        y_shift = self.YDir.UVectECF.get_array() * self.YDir.SampleSpacing
        # order convention
        x_offset = numpy.array(
            [self.XDir.FirstLine, self.XDir.FirstLine,
             self.XDir.FirstLine + self.XDir.NumLines - 1, self.XDir.FirstLine + self.XDir.NumLines - 1])
        y_offset = numpy.array(
            [self.YDir.FirstSample, self.YDir.FirstSample + self.YDir.NumSamples - 1,
             self.YDir.FirstSample + self.YDir.NumSamples - 1, self.YDir.FirstSample])
        corners = numpy.zeros((4, 3), dtype=numpy.float64)
        for i in range(4):
            corners[i, :] = \
                ecf_ref + x_shift*(x_offset[i] - self.RefPt.Line) + y_shift*(y_offset[i] - self.RefPt.Sample)
        return corners


class AreaType(Serializable):
    """
    The collection area.
    """

    _fields = ('Corner', 'Plane')
    _required = ('Corner', )
    _collections_tags = {
        'Corner': {'array': True, 'child_tag': 'ACP'}, }
    # descriptors
    Corner = SerializableCPArrayDescriptor(
        'Corner', LatLonHAECornerRestrictionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The collection area corner point definition array.'
    )  # type: Union[SerializableCPArray, List[LatLonHAECornerRestrictionType]]
    Plane = SerializableDescriptor(
        'Plane', ReferencePlaneType, _required, strict=DEFAULT_STRICT,
        docstring='A rectangular area in a geo-located display plane.')  # type: ReferencePlaneType

    def __init__(
            self,
            Corner: Union[SerializableCPArray, List[LatLonHAECornerRestrictionType], numpy.ndarray, list, tuple] = None,
            Plane: ReferencePlaneType = None,
            **kwargs):
        """

        Parameters
        ----------
        Corner : SerializableCPArray|List[LatLonHAECornerRestrictionType]|numpy.ndarray|list|tuple
        Plane : ReferencePlaneType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Corner = Corner
        self.Plane = Plane
        super(AreaType, self).__init__(**kwargs)
        self.derive()

    def _derive_corner_from_plane(self):
        # try to define the corner points - for SICD 0.5.
        if self.Corner is not None:
            return  # nothing to be done
        if self.Plane is None:
            return  # nothing to derive from
        corners = self.Plane.get_ecf_corner_array()
        self.Corner = [
            LatLonHAECornerRestrictionType(**{'Lat': entry[0], 'Lon': entry[1], 'HAE': entry[2], 'index': i+1})
            for i, entry in enumerate(geocoords.ecf_to_geodetic(corners))]

    def derive(self):
        """
        Derive the corner points from the plane, if necessary.

        Returns
        -------
        None
        """

        self._derive_corner_from_plane()


class RadarCollectionType(Serializable):
    """The Radar Collection Type"""
    _fields = (
        'TxFrequency', 'RefFreqIndex', 'Waveform', 'TxPolarization', 'TxSequence', 'RcvChannels', 'Area', 'Parameters')
    _required = ('TxFrequency', 'TxPolarization', 'RcvChannels')
    _collections_tags = {
        'Waveform': {'array': True, 'child_tag': 'WFParameters'},
        'TxSequence': {'array': True, 'child_tag': 'TxStep'},
        'RcvChannels': {'array': True, 'child_tag': 'ChanParameters'},
        'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    TxFrequency = SerializableDescriptor(
        'TxFrequency', TxFrequencyType, _required, strict=DEFAULT_STRICT,
        docstring='The transmit frequency range.')  # type: TxFrequencyType
    RefFreqIndex = IntegerDescriptor(
        'RefFreqIndex', _required, strict=DEFAULT_STRICT,
        docstring='The reference frequency index, if applicable. If present and non-zero, '
                  'all (most) RF frequency values are expressed as offsets from a reference '
                  'frequency.')  # type: int
    Waveform = SerializableArrayDescriptor(
        'Waveform', WaveformParametersType, _collections_tags, _required,
        strict=DEFAULT_STRICT, minimum_length=1,
        docstring='Transmit and receive demodulation waveform parameters.'
    )  # type: Union[SerializableArray, List[WaveformParametersType]]
    TxPolarization = StringEnumDescriptor(
        'TxPolarization', POLARIZATION1_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='The transmit polarization.')  # type: str
    TxSequence = SerializableArrayDescriptor(
        'TxSequence', TxStepType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=1,
        docstring='The transmit sequence parameters array. If present, indicates the transmit signal steps through '
                  'a repeating sequence of waveforms and/or polarizations. '
                  'One step per Inter-Pulse Period.')  # type: Union[None, SerializableArray, List[TxStepType]]
    RcvChannels = SerializableArrayDescriptor(
        'RcvChannels', ChanParametersType, _collections_tags,
        _required, strict=DEFAULT_STRICT, minimum_length=1,
        docstring='Receive data channel parameters.')  # type: Union[SerializableArray, List[ChanParametersType]]
    Area = SerializableDescriptor(
        'Area', AreaType, _required, strict=DEFAULT_STRICT,
        docstring='The imaged area covered by the collection.')  # type: AreaType
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='A parameters collections.')  # type: ParametersCollection

    def __init__(
            self,
            TxFrequency: TxFrequencyType = None,
            RefFreqIndex: Optional[int] = None,
            Waveform: Union[None, SerializableArray, List[WaveformParametersType]] = None,
            TxPolarization: str = None,
            TxSequence: Union[None, SerializableArray, List[TxStepType]] = None,
            RcvChannels: Union[SerializableArray, List[ChanParametersType]] = None,
            Area: Optional[AreaType] = None,
            Parameters: Union[None, ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        TxFrequency : TxFrequencyType|numpy.ndarray|list|tuple
        RefFreqIndex : int
        Waveform : SerializableArray|List[WaveformParametersType]
        TxPolarization : str
        TxSequence : SerializableArray|List[TxStepType]
        RcvChannels : SerializableArray|List[ChanParametersType]
        Area : AreaType
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxFrequency = TxFrequency
        self.RefFreqIndex = RefFreqIndex
        self.Waveform = Waveform
        self.TxPolarization = TxPolarization
        self.TxSequence = TxSequence
        self.RcvChannels = RcvChannels
        self.Area = Area
        self.Parameters = Parameters
        super(RadarCollectionType, self).__init__(**kwargs)

    def derive(self):
        """
        Populates derived data in RadarCollection. Expected to be called by `SICD` parent.

        Returns
        -------
        None
        """

        self._derive_tx_polarization()
        if self.Area is not None:
            self.Area.derive()
        if self.Waveform is not None:
            for entry in self.Waveform:
                entry.derive()
        self._derive_tx_frequency()  # call after waveform entry derive call
        self._derive_wf_params()

    def _derive_tx_polarization(self):
        def check_sequence():
            unique_entries = set(entry.TxPolarization for entry in self.TxSequence)
            if len(unique_entries) == 1:
                self.TxPolarization = self.TxSequence[0].TxPolarization
            else:
                self.TxPolarization = 'SEQUENCE'

        # TxPolarization was optional prior to SICD 1.0. It may need to be derived.
        if self.TxSequence is not None:
            check_sequence()
            return
        if self.TxPolarization is not None:
            return  # nothing to be done

        if self.RcvChannels is None:
            return  # nothing to derive from

        if len(self.RcvChannels) > 1:
            # TxSequence may need to be derived from RCvChannels, for SICD before 1.0 or poorly formed
            if self.TxSequence is not None or self.RcvChannels is None or len(self.RcvChannels) < 2:
                return

            tx_pols = list(chan_param.get_transmit_polarization() for chan_param in self.RcvChannels)

            if len(tx_pols) > 1:
                self.TxSequence = [TxStepType(index=i+1, TxPolarization=tx_pol) for i, tx_pol in enumerate(tx_pols)]
                check_sequence()
            else:
                self.TxPolarization = tx_pols[0]
        else:
            self.TxPolarization = self.RcvChannels[0].get_transmit_polarization()

    def _derive_tx_frequency(self):
        if self.Waveform is None or self.Waveform.size == 0:
            return  # nothing to be done
        if not(self.TxFrequency is None or self.TxFrequency.Min is None or self.TxFrequency.Max is None):
            return  # no need to do anything

        if self.TxFrequency is None:
            self.TxFrequency = TxFrequencyType()
        if self.TxFrequency.Min is None:
            self.TxFrequency.Min = min(
                entry.TxFreqStart for entry in self.Waveform if entry.TxFreqStart is not None)
        if self.TxFrequency.Max is None:
            self.TxFrequency.Max = max(
                (entry.TxFreqStart + entry.TxRFBandwidth) for entry in self.Waveform if
                entry.TxFreqStart is not None and entry.TxRFBandwidth is not None)

    def _derive_wf_params(self):
        if self.TxFrequency is None or self.TxFrequency.Min is None or self.TxFrequency.Max is None:
            return  # nothing that we can do
        if self.Waveform is None or self.Waveform.size != 1:
            return  # nothing to be done

        entry = self.Waveform[0]  # only true for single waveform definition
        if entry.TxFreqStart is None:
            entry.TxFreqStart = self.TxFrequency.Min
        if entry.TxRFBandwidth is None:
            entry.TxRFBandwidth = self.TxFrequency.Max - self.TxFrequency.Min

    def _apply_reference_frequency(self, reference_frequency):
        """
        If the reference frequency is used, adjust the necessary fields accordingly.
        Expected to be called by `SICD` parent.

        Parameters
        ----------
        reference_frequency : float
            The reference frequency.

        Returns
        -------
        None
        """

        if self.TxFrequency is not None:
            # noinspection PyProtectedMember
            self.TxFrequency._apply_reference_frequency(reference_frequency)
        if self.Waveform is not None:
            for entry in self.Waveform:
                # noinspection PyProtectedMember
                entry._apply_reference_frequency(reference_frequency)
        self.RefFreqIndex = 0

    def get_polarization_abbreviation(self):
        """
        Gets the polarization collection abbreviation for the suggested name.

        Returns
        -------
        str
        """

        if self.RcvChannels is None:
            pol_count = 0
        else:
            pol_count = len(self.RcvChannels)
        if pol_count == 1:
            return 'S'
        elif pol_count == 2:
            return 'D'
        elif pol_count == 3:
            return 'T'
        elif pol_count > 3:
            return 'Q'
        else:
            return 'U'

    def _check_frequency(self):
        # type: () -> bool

        if self.RefFreqIndex is not None:
            return True

        if self.TxFrequency is not None and self.TxFrequency.Min is not None \
                and self.TxFrequency.Min <= 0:
            self.log_validity_error(
                'TxFrequency.Min is negative, but RefFreqIndex is not populated.')
            return False
        return True

    def _check_tx_sequence(self):
        # type: () -> bool

        cond = True
        if self.TxPolarization == 'SEQUENCE' and self.TxSequence is None:
            self.log_validity_error(
                'TxPolarization is populated as "SEQUENCE", but TxSequence is not populated.')
            cond = False
        if self.TxSequence is not None:
            if self.TxPolarization != 'SEQUENCE':
                self.log_validity_error(
                    'TxSequence is populated, but TxPolarization is populated as {}'.format(self.TxPolarization))
                cond = False
            tx_pols = list(set([entry.TxPolarization for entry in self.TxSequence]))
            if len(tx_pols) == 1:
                self.log_validity_error(
                    'TxSequence is populated, but the only unique TxPolarization '
                    'among the entries is {}'.format(tx_pols[0]))
                cond = False
        return cond

    def _check_waveform_parameters(self):
        """
        Validate the waveform parameters for consistency.

        Returns
        -------
        bool
        """

        def validate_entry(index, waveform):
            # type: (int, WaveformParametersType) -> bool
            this_cond = True
            try:
                if abs(waveform.TxRFBandwidth/(waveform.TxPulseLength*waveform.TxFMRate) - 1) > 1e-3:
                    self.log_validity_error(
                        'The TxRFBandwidth, TxPulseLength, and TxFMRate parameters of Waveform '
                        'entry {} are inconsistent'.format(index+1))
                    this_cond = False
            except (AttributeError, ValueError, TypeError):
                pass

            if waveform.RcvDemodType == 'CHIRP' and waveform.RcvFMRate != 0:
                self.log_validity_error(
                    'RcvDemodType == "CHIRP" and RcvFMRate != 0 in Waveform entry {}'.format(index+1))
                this_cond = False
            if waveform.RcvDemodType == 'STRETCH' and \
                    waveform.RcvFMRate is not None and waveform.TxFMRate is not None and \
                    abs(waveform.RcvFMRate/waveform.TxFMRate - 1) > 1e-3:
                self.log_validity_warning(
                    'RcvDemodType = "STRETCH", RcvFMRate = {}, and TxFMRate = {} in '
                    'Waveform entry {}. The RcvFMRate and TxFMRate should very likely '
                    'be the same.'.format(waveform.RcvFMRate, waveform.TxFMRate, index+1))

            if self.RefFreqIndex is None:
                if waveform.TxFreqStart is not None and waveform.TxFreqStart <= 0:
                    self.log_validity_error(
                        'TxFreqStart is negative in Waveform entry {}, but RefFreqIndex '
                        'is not populated.'.format(index+1))
                    this_cond = False
                if waveform.RcvFreqStart is not None and waveform.RcvFreqStart <= 0:
                    self.log_validity_error(
                        'RcvFreqStart is negative in Waveform entry {}, but RefFreqIndex '
                        'is not populated.'.format(index + 1))
                    this_cond = False
            if waveform.TxPulseLength is not None and waveform.RcvWindowLength is not None and \
                    waveform.TxPulseLength > waveform.RcvWindowLength:
                self.log_validity_error(
                    'TxPulseLength ({}) is longer than RcvWindowLength ({}) in '
                    'Waveform entry {}'.format(waveform.TxPulseLength, waveform.RcvWindowLength, index+1))
                this_cond = False
            if waveform.RcvIFBandwidth is not None and waveform.ADCSampleRate is not None and \
                    waveform.RcvIFBandwidth > waveform.ADCSampleRate:
                self.log_validity_error(
                    'RcvIFBandwidth ({}) is longer than ADCSampleRate ({}) in '
                    'Waveform entry {}'.format(waveform.RcvIFBandwidth, waveform.ADCSampleRate, index+1))
                this_cond = False
            if waveform.RcvDemodType is not None and waveform.RcvDemodType == 'CHIRP' \
                    and waveform.TxRFBandwidth is not None and waveform.ADCSampleRate is not None \
                    and (waveform.TxRFBandwidth > waveform.ADCSampleRate):
                self.log_validity_error(
                    'RcvDemodType is "CHIRP" and TxRFBandwidth ({}) is larger than ADCSampleRate ({}) '
                    'in Waveform entry {}'.format(waveform.TxRFBandwidth, waveform.ADCSampleRate, index+1))
                this_cond = False
            if waveform.RcvWindowLength is not None and waveform.TxPulseLength is not None and \
                    waveform.TxFMRate is not None and waveform.RcvFreqStart is not None and \
                    waveform.TxFreqStart is not None and waveform.TxRFBandwidth is not None:
                freq_tol = (waveform.RcvWindowLength - waveform.TxPulseLength)*waveform.TxFMRate
                if waveform.RcvFreqStart >= waveform.TxFreqStart + waveform.TxRFBandwidth + freq_tol:
                    self.log_validity_error(
                        'RcvFreqStart ({}), TxFreqStart ({}), and TxRfBandwidth ({}) parameters are inconsistent '
                        'in Waveform entry {}'.format(
                            waveform.RcvFreqStart, waveform.TxFreqStart, waveform.TxRFBandwidth, index + 1))
                    this_cond = False
                if waveform.RcvFreqStart <= waveform.TxFreqStart - freq_tol:
                    self.log_validity_error(
                        'RcvFreqStart ({}) and TxFreqStart ({}) parameters are inconsistent '
                        'in Waveform entry {}'.format(waveform.RcvFreqStart, waveform.TxFreqStart, index + 1))
                    this_cond = False

            return this_cond

        if self.Waveform is None or len(self.Waveform) < 1:
            return True

        cond = True
        # fetch min/max TxFreq observed
        wf_min_freq = None
        wf_max_freq = None
        for entry in self.Waveform:
            freq_start = entry.TxFreqStart
            freq_bw = entry.TxRFBandwidth
            if freq_start is not None:
                wf_min_freq = freq_start if wf_min_freq is None else \
                    min(wf_min_freq, freq_start)
                if entry.TxRFBandwidth is not None:
                    wf_max_freq = freq_start + freq_bw if wf_max_freq is None else \
                        max(wf_max_freq, freq_start + freq_bw)

        if wf_min_freq is not None and self.TxFrequency is not None and self.TxFrequency.Min is not None:
            if abs(self.TxFrequency.Min/wf_min_freq - 1) > 1e-3:
                self.log_validity_error(
                    'The stated TxFrequency.Min is {}, but the minimum populated in a '
                    'Waveform entry is {}'.format(self.TxFrequency.Min, wf_min_freq))
                cond = False
        if wf_max_freq is not None and self.TxFrequency is not None and self.TxFrequency.Max is not None:
            if abs(self.TxFrequency.Max/wf_max_freq - 1) > 1e-3:
                self.log_validity_error(
                    'The stated TxFrequency.Max is {}, but the maximum populated in a '
                    'Waveform entry is {}'.format(self.TxFrequency.Max, wf_max_freq))
                cond = False
        for t_index, t_waveform in enumerate(self.Waveform):
            cond &= validate_entry(t_index, t_waveform)
        return cond

    def _basic_validity_check(self):
        valid = super(RadarCollectionType, self)._basic_validity_check()
        valid &= self._check_frequency()
        valid &= self._check_tx_sequence()
        valid &= self._check_waveform_parameters()
        return valid

    def version_required(self):
        """
        What SICD version is required?

        Returns
        -------
        tuple
        """

        requires = (1, 1, 0)
        if self.RcvChannels is None:
            return requires

        for entry in self.RcvChannels:
            requires = max(requires, entry.version_required())
        return requires




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class STREOBType(TREElement):
    def __init__(self, value):
        super(STREOBType, self).__init__()
        self.add_field('ST_ID', 's', 60, value)
        self.add_field('N_MATES', 's', 1, value)
        self.add_field('MATE_INSTANCE', 's', 1, value)
        self.add_field('B_CONV', 's', 5, value)
        self.add_field('E_CONV', 's', 5, value)
        self.add_field('B_ASYM', 's', 5, value)
        self.add_field('E_ASYM', 's', 5, value)
        self.add_field('B_BIE', 's', 6, value)
        self.add_field('E_BIE', 's', 6, value)


class STREOB(TREExtension):
    _tag_value = 'STREOB'
    _data_type = STREOBType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class MENSRBType(TREElement):
    def __init__(self, value):
        super(MENSRBType, self).__init__()
        self.add_field('ACFT_LOC', 's', 25, value)
        self.add_field('ACFT_LOC_ACCY', 's', 6, value)
        self.add_field('ACFT_ALT', 's', 6, value)
        self.add_field('RP_LOC', 's', 25, value)
        self.add_field('RP_LOC_ACCY', 's', 6, value)
        self.add_field('RP_ELV', 's', 6, value)
        self.add_field('OF_PC_R', 's', 7, value)
        self.add_field('OF_PC_A', 's', 7, value)
        self.add_field('COSGRZ', 's', 7, value)
        self.add_field('RGCRP', 's', 7, value)
        self.add_field('RLMAP', 's', 1, value)
        self.add_field('RP_ROW', 's', 5, value)
        self.add_field('RP_COL', 's', 5, value)
        self.add_field('C_R_NC', 's', 10, value)
        self.add_field('C_R_EC', 's', 10, value)
        self.add_field('C_R_DC', 's', 10, value)
        self.add_field('C_AZ_NC', 's', 9, value)
        self.add_field('C_AZ_EC', 's', 9, value)
        self.add_field('C_AZ_DC', 's', 9, value)
        self.add_field('C_AL_NC', 's', 9, value)
        self.add_field('C_AL_EC', 's', 9, value)
        self.add_field('C_AL_DC', 's', 9, value)
        self.add_field('TOTAL_TILES_COLS', 's', 3, value)
        self.add_field('TOTAL_TILES_ROWS', 's', 5, value)


class MENSRB(TREExtension):
    _tag_value = 'MENSRB'
    _data_type = MENSRBType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CSPROAType(TREElement):
    def __init__(self, value):
        super(CSPROAType, self).__init__()
        self.add_field('RESERVED_0', 's', 12, value)
        self.add_field('RESERVED_1', 's', 12, value)
        self.add_field('RESERVED_2', 's', 12, value)
        self.add_field('RESERVED_3', 's', 12, value)
        self.add_field('RESERVED_4', 's', 12, value)
        self.add_field('RESERVED_5', 's', 12, value)
        self.add_field('RESERVED_6', 's', 12, value)
        self.add_field('RESERVED_7', 's', 12, value)
        self.add_field('RESERVED_8', 's', 12, value)
        self.add_field('BWC', 's', 12, value)


class CSPROA(TREExtension):
    _tag_value = 'CSPROA'
    _data_type = CSPROAType



"""
The TxRcv type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import FloatDescriptor, StringDescriptor, \
    StringEnumDescriptor, SerializableListDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import POLARIZATION_TYPE


class TxWFParametersType(Serializable):
    """
    Parameters that describe a Transmit Waveform.
    """

    _fields = (
        'Identifier', 'PulseLength', 'RFBandwidth', 'FreqCenter', 'LFMRate',
        'Polarization', 'Power')
    _required = (
        'Identifier', 'PulseLength', 'RFBandwidth', 'FreqCenter', 'Polarization')
    _numeric_format = {
        'PulseLength': FLOAT_FORMAT, 'RFBandwidth': FLOAT_FORMAT, 'FreqCenter': FLOAT_FORMAT,
        'LFMRate': FLOAT_FORMAT, 'Power': FLOAT_FORMAT}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this Transmit '
                  'Waveform.')  # type: str
    PulseLength = FloatDescriptor(
        'PulseLength', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Length of transmitted pulse, '
                  'in seconds.')  # type: float
    RFBandwidth = FloatDescriptor(
        'RFBandwidth', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Bandwidth of transmitted pulse, '
                  'in Hz.')  # type: float
    FreqCenter = FloatDescriptor(
        'FreqCenter', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Center frequency of the transmitted waveform, '
                  'in Hz.')  # type: float
    LFMRate = FloatDescriptor(
        'LFMRate', _required, strict=DEFAULT_STRICT,
        docstring='Chirp rate of transmitted pulse if LFM, '
                  'in Hz/s.')  # type: Union[None, float]
    Polarization = StringEnumDescriptor(
        'Polarization', POLARIZATION_TYPE, _required, strict=DEFAULT_STRICT,
        docstring='The transmit polarization mode.')  # type: str
    Power = FloatDescriptor(
        'Power', _required, strict=DEFAULT_STRICT,
        docstring='Peak transmitted power at the interface to the antenna '
                  'in dBW.')  # type: Union[None, float]

    def __init__(self, Identifier=None, PulseLength=None, RFBandwidth=None,
                 FreqCenter=None, LFMRate=None, Polarization=None, Power=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        PulseLength : float
        RFBandwidth : float
        FreqCenter : float
        LFMRate : None|float
        Polarization : str
        Power : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.PulseLength = PulseLength
        self.RFBandwidth = RFBandwidth
        self.FreqCenter = FreqCenter
        self.LFMRate = LFMRate
        self.Polarization = Polarization
        self.Power = Power
        super(TxWFParametersType, self).__init__(**kwargs)


class RcvParametersType(Serializable):
    """
    Parameters that describe a Receive configuration.
    """

    _fields = (
        'Identifier', 'WindowLength', 'SampleRate', 'IFFilterBW', 'FreqCenter',
        'LFMRate', 'Polarization', 'PathGain')
    _required = (
        'Identifier', 'WindowLength', 'SampleRate', 'IFFilterBW', 'FreqCenter',
        'Polarization')
    _numeric_format = {
        'WindowLength': FLOAT_FORMAT, 'SampleRate': FLOAT_FORMAT, 'IFFilterBW': FLOAT_FORMAT,
        'FreqCenter': FLOAT_FORMAT, 'LFMRate': FLOAT_FORMAT, 'PathGain': FLOAT_FORMAT}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this Receive '
                  'configuration.')  # type: str
    WindowLength = FloatDescriptor(
        'WindowLength', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Length of the receive window, in seconds.')  # type: float
    SampleRate = FloatDescriptor(
        'SampleRate', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Rate at which the signal in the receive window is sampled, '
                  'in Hz.')  # type: float
    IFFilterBW = FloatDescriptor(
        'IFFilterBW', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Bandwidth of the anti-aliasing filter prior to '
                  'sampling.')  # type: float
    FreqCenter = FloatDescriptor(
        'FreqCenter', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Center frequency of the demodulation signal, '
                  'in Hz.')  # type: float
    LFMRate = FloatDescriptor(
        'LFMRate', _required, strict=DEFAULT_STRICT,
        docstring='Chirp rate of the demodulation signal if LFM, '
                  'in Hz/s.')  # type: Union[None, float]
    Polarization = StringEnumDescriptor(
        'Polarization', POLARIZATION_TYPE, _required, strict=DEFAULT_STRICT,
        docstring='The receive polarization mode.')  # type: str
    PathGain = FloatDescriptor(
        'PathGain', _required, strict=DEFAULT_STRICT,
        docstring='Receiver gain from the antenna interface to the ADC, '
                  'in dB.')  # type: Union[None, float]

    def __init__(self, Identifier=None, WindowLength=None, SampleRate=None, IFFilterBW=None,
                 FreqCenter=None, LFMRate=None, Polarization=None, PathGain=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        WindowLength : float
        SampleRate : float
        IFFilterBW : float
        FreqCenter : float
        LFMRate : None|float
        Polarization : str
        PathGain : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.WindowLength = WindowLength
        self.SampleRate = SampleRate
        self.IFFilterBW = IFFilterBW
        self.FreqCenter = FreqCenter
        self.LFMRate = LFMRate
        self.Polarization = Polarization
        self.PathGain = PathGain
        super(RcvParametersType, self).__init__(**kwargs)


class TxRcvType(Serializable):
    """
    Parameters that describe the transmitted waveform(s) and receiver
    configurations used in the collection
    """

    _fields = ('NumTxWFs', 'TxWFParameters', 'NumRcvs', 'RcvParameters')
    _required = _fields
    _collections_tags = {
        'TxWFParameters': {'array': False, 'child_tag': 'TxWFParameters'},
        'RcvParameters': {'array': False, 'child_tag': 'RcvParameters'}}
    # descriptors
    TxWFParameters = SerializableListDescriptor(
        'TxWFParameters', TxWFParametersType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe a Transmit Waveform.')  # type: List[TxWFParametersType]
    RcvParameters = SerializableListDescriptor(
        'RcvParameters', RcvParametersType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe a Receive configuration.')  # type: List[RcvParametersType]

    def __init__(self, TxWFParameters=None, RcvParameters=None, **kwargs):
        """

        Parameters
        ----------
        TxWFParameters : List[TxWFParametersType]
        RcvParameters : List[RcvParametersType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxWFParameters = TxWFParameters
        self.RcvParameters = RcvParameters
        super(TxRcvType, self).__init__(**kwargs)

    @property
    def NumTxWFs(self):
        """
        int: The number of transmit waveforms used.
        """

        if self.TxWFParameters is None:
            return 0
        return len(self.TxWFParameters)

    @property
    def NumRcvs(self):
        """
        int: The number of receive configurations used.
        """

        if self.RcvParameters is None:
            return 0
        return len(self.RcvParameters)



"""
The ErrorStatisticsType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, Optional, Dict
from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import StringEnumDescriptor, FloatDescriptor, \
    SerializableDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import ErrorDecorrFuncType


class CompositeSCPErrorType(Serializable):
    """
    Composite error statistics for the Scene Center Point. Slant plane range *(Rg)*
    and azimuth *(Az)* error statistics. Slant plane defined at *SCP COA*.
    """
    _fields = ('Rg', 'Az', 'RgAz')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # descriptors
    Rg = FloatDescriptor(
        'Rg', _required, strict=DEFAULT_STRICT,
        docstring='Estimated range error standard deviation.')  # type: float
    Az = FloatDescriptor(
        'Az', _required, strict=DEFAULT_STRICT,
        docstring='Estimated azimuth error standard deviation.')  # type: float
    RgAz = FloatDescriptor(
        'RgAz', _required, strict=DEFAULT_STRICT,
        docstring='Estimated range and azimuth error correlation coefficient.')  # type: float

    def __init__(
            self,
            Rg: float = None,
            Az: float = None,
            RgAz: float = None,
            **kwargs):
        """

        Parameters
        ----------
        Rg : float
        Az : float
        RgAz : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Rg, self.Az, self.RgAz = Rg, Az, RgAz
        super(CompositeSCPErrorType, self).__init__(**kwargs)


class CorrCoefsType(Serializable):
    """Correlation Coefficient parameters."""
    _fields = (
        'P1P2', 'P1P3', 'P1V1', 'P1V2', 'P1V3', 'P2P3', 'P2V1', 'P2V2', 'P2V3',
        'P3V1', 'P3V2', 'P3V3', 'V1V2', 'V1V3', 'V2V3')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # descriptors
    P1P2 = FloatDescriptor(
        'P1P2', _required, strict=DEFAULT_STRICT, docstring='`P1` and `P2` correlation coefficient.')  # type: float
    P1P3 = FloatDescriptor(
        'P1P3', _required, strict=DEFAULT_STRICT, docstring='`P1` and `P3` correlation coefficient.')  # type: float
    P1V1 = FloatDescriptor(
        'P1V1', _required, strict=DEFAULT_STRICT, docstring='`P1` and `V1` correlation coefficient.')  # type: float
    P1V2 = FloatDescriptor(
        'P1V2', _required, strict=DEFAULT_STRICT, docstring='`P1` and `V2` correlation coefficient.')  # type: float
    P1V3 = FloatDescriptor(
        'P1V3', _required, strict=DEFAULT_STRICT, docstring='`P1` and `V3` correlation coefficient.')  # type: float
    P2P3 = FloatDescriptor(
        'P2P3', _required, strict=DEFAULT_STRICT, docstring='`P2` and `P3` correlation coefficient.')  # type: float
    P2V1 = FloatDescriptor(
        'P2V1', _required, strict=DEFAULT_STRICT, docstring='`P2` and `V1` correlation coefficient.')  # type: float
    P2V2 = FloatDescriptor(
        'P2V2', _required, strict=DEFAULT_STRICT, docstring='`P2` and `V2` correlation coefficient.')  # type: float
    P2V3 = FloatDescriptor(
        'P2V3', _required, strict=DEFAULT_STRICT, docstring='`P2` and `V3` correlation coefficient.')  # type: float
    P3V1 = FloatDescriptor(
        'P3V1', _required, strict=DEFAULT_STRICT, docstring='`P3` and `V1` correlation coefficient.')  # type: float
    P3V2 = FloatDescriptor(
        'P3V2', _required, strict=DEFAULT_STRICT, docstring='`P3` and `V2` correlation coefficient.')  # type: float
    P3V3 = FloatDescriptor(
        'P3V3', _required, strict=DEFAULT_STRICT, docstring='`P3` and `V3` correlation coefficient.')  # type: float
    V1V2 = FloatDescriptor(
        'V1V2', _required, strict=DEFAULT_STRICT, docstring='`V1` and `V2` correlation coefficient.')  # type: float
    V1V3 = FloatDescriptor(
        'V1V3', _required, strict=DEFAULT_STRICT, docstring='`V1` and `V3` correlation coefficient.')  # type: float
    V2V3 = FloatDescriptor(
        'V2V3', _required, strict=DEFAULT_STRICT, docstring='`V2` and `V3` correlation coefficient.')  # type: float

    def __init__(
            self,
            P1P2: float = None,
            P1P3: float = None,
            P1V1: float = None,
            P1V2: float = None,
            P1V3: float = None,
            P2P3: float = None,
            P2V1: float = None,
            P2V2: float = None,
            P2V3: float = None,
            P3V1: float = None,
            P3V2: float = None,
            P3V3: float = None,
            V1V2: float = None,
            V1V3: float = None,
            V2V3: float = None,
            **kwargs):
        """

        Parameters
        ----------
        P1P2 : float
        P1P3 : float
        P1V1 : float
        P1V2 : float
        P1V3 : float
        P2P3 : float
        P2V1 : float
        P2V2 : float
        P2V3 : float
        P3V1 : float
        P3V2 : float
        P3V3 : float
        V1V2 : float
        V1V3 : float
        V2V3 : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.P1P2, self.P1P3, self.P1V1, self.P1V2, self.P1V3 = P1P2, P1P3, P1V1, P1V2, P1V3
        self.P2P3, self.P2V1, self.P2V2, self.P2V3 = P2P3, P2V1, P2V2, P2V3
        self.P3V1, self.P3V2, self.P3V3 = P3V1, P3V2, P3V3
        self.V1V2, self.V1V3 = V1V2, V1V3
        self.V2V3 = V2V3
        super(CorrCoefsType, self).__init__(**kwargs)


class PosVelErrType(Serializable):
    """
    Position and velocity error statistics for the radar platform.
    """

    _fields = ('Frame', 'P1', 'P2', 'P3', 'V1', 'V2', 'V3', 'CorrCoefs', 'PositionDecorr')
    _required = ('Frame', 'P1', 'P2', 'P3', 'V1', 'V2', 'V3')
    _numeric_format = {
        'P1': FLOAT_FORMAT, 'P2': FLOAT_FORMAT, 'P3': FLOAT_FORMAT,
        'V1': FLOAT_FORMAT, 'V2': FLOAT_FORMAT, 'V3': FLOAT_FORMAT}
    # class variables
    _FRAME_VALUES = ('ECF', 'RIC_ECF', 'RIC_ECI')
    # descriptors
    Frame = StringEnumDescriptor(
        'Frame', _FRAME_VALUES, _required, strict=True,
        docstring='Coordinate frame used for expressing P,V errors statistics. Note: '
                  '*RIC = Radial, In-Track, Cross-Track*, where radial is defined to be from earth center through '
                  'the platform position.')  # type: str
    P1 = FloatDescriptor(
        'P1', _required, strict=DEFAULT_STRICT, docstring='Position coordinate 1 standard deviation.')  # type: float
    P2 = FloatDescriptor(
        'P2', _required, strict=DEFAULT_STRICT, docstring='Position coordinate 2 standard deviation.')  # type: float
    P3 = FloatDescriptor(
        'P3', _required, strict=DEFAULT_STRICT, docstring='Position coordinate 3 standard deviation.')  # type: float
    V1 = FloatDescriptor(
        'V1', _required, strict=DEFAULT_STRICT, docstring='Velocity coordinate 1 standard deviation.')  # type: float
    V2 = FloatDescriptor(
        'V2', _required, strict=DEFAULT_STRICT, docstring='Velocity coordinate 2 standard deviation.')  # type: float
    V3 = FloatDescriptor(
        'V3', _required, strict=DEFAULT_STRICT, docstring='Velocity coordinate 3 standard deviation.')  # type: float
    CorrCoefs = SerializableDescriptor(
        'CorrCoefs', CorrCoefsType, _required, strict=DEFAULT_STRICT,
        docstring='Correlation Coefficient parameters.')  # type: Optional[CorrCoefsType]
    PositionDecorr = SerializableDescriptor(
        'PositionDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Platform position error decorrelation function.')  # type: Optional[ErrorDecorrFuncType]

    def __init__(
            self,
            Frame: str = None,
            P1: float = None,
            P2: float = None,
            P3: float = None,
            V1: float = None,
            V2: float = None,
            V3: float = None,
            CorrCoefs: Optional[CorrCoefsType] = None,
            PositionDecorr: Optional[ErrorDecorrFuncType] = None,
            **kwargs):
        """

        Parameters
        ----------
        Frame : str
        P1 : float
        P2 : float
        P3 : float
        V1 : float
        V2 : float
        V3 : float
        CorrCoefs : None|CorrCoefsType
        PositionDecorr : None|ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Frame = Frame
        self.P1, self.P2, self.P3 = P1, P2, P3
        self.V1, self.V2, self.V3 = V1, V2, V3
        self.CorrCoefs, self.PositionDecorr = CorrCoefs, PositionDecorr
        super(PosVelErrType, self).__init__(**kwargs)


class RadarSensorErrorType(Serializable):
    """Radar sensor error statistics."""
    _fields = ('RangeBias', 'ClockFreqSF', 'TransmitFreqSF', 'RangeBiasDecorr')
    _required = ('RangeBias', )
    _numeric_format = {'RangeBias': FLOAT_FORMAT, 'ClockFreqSF': FLOAT_FORMAT, 'TransmitFreqSF': FLOAT_FORMAT}
    # descriptors
    RangeBias = FloatDescriptor(
        'RangeBias', _required, strict=DEFAULT_STRICT,
        docstring='Range bias error standard deviation.')  # type: float
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT,
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    TransmitFreqSF = FloatDescriptor(
        'TransmitFreqSF', _required, strict=DEFAULT_STRICT,
        docstring='Transmit frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    RangeBiasDecorr = SerializableDescriptor(
        'RangeBiasDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Range bias decorrelation rate.')  # type: ErrorDecorrFuncType

    def __init__(
            self,
            RangeBias: float = None,
            ClockFreqSF: Optional[float] = None,
            TransmitFreqSF: Optional[float] = None,
            RangeBiasDecorr: Optional[ErrorDecorrFuncType] = None,
            **kwargs):
        """

        Parameters
        ----------
        RangeBias : float
        ClockFreqSF : None|float
        TransmitFreqSF : None|float
        RangeBiasDecorr : None|ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RangeBias, self.ClockFreqSF, self.TransmitFreqSF = RangeBias, ClockFreqSF, TransmitFreqSF
        self.RangeBiasDecorr = RangeBiasDecorr
        super(RadarSensorErrorType, self).__init__(**kwargs)


class TropoErrorType(Serializable):
    """Troposphere delay error statistics."""
    _fields = ('TropoRangeVertical', 'TropoRangeSlant', 'TropoRangeDecorr')
    _required = ()
    _numeric_format = {'TropoRangeVertical': FLOAT_FORMAT, 'TropoRangeSlant': FLOAT_FORMAT}
    # descriptors
    TropoRangeVertical = FloatDescriptor(
        'TropoRangeVertical', _required, strict=DEFAULT_STRICT,
        docstring='Troposphere two-way delay error for normal incidence standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    TropoRangeSlant = FloatDescriptor(
        'TropoRangeSlant', _required, strict=DEFAULT_STRICT,
        docstring='Troposphere two-way delay error for the *SCP* line of sight at *COA* standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    TropoRangeDecorr = SerializableDescriptor(
        'TropoRangeDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Troposphere range error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(
            self,
            TropoRangeVertical: Optional[float] = None,
            TropoRangeSlant: Optional[float] = None,
            TropoRangeDecorr: Optional[ErrorDecorrFuncType] = None,
            **kwargs):
        """

        Parameters
        ----------
        TropoRangeVertical : None|float
        TropoRangeSlant : None|float
        TropoRangeDecorr : None|ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TropoRangeVertical = TropoRangeVertical
        self.TropoRangeSlant = TropoRangeSlant
        self.TropoRangeDecorr = TropoRangeDecorr
        super(TropoErrorType, self).__init__(**kwargs)


class IonoErrorType(Serializable):
    """Ionosphere delay error statistics."""
    _fields = ('IonoRangeVertical', 'IonoRangeRateVertical', 'IonoRgRgRateCC', 'IonoRangeVertDecorr')
    _required = ('IonoRgRgRateCC', )
    _numeric_format = {
        'IonoRangeVertical': FLOAT_FORMAT, 'IonoRangeRateVertical': FLOAT_FORMAT, 'IonoRgRgRateCC': FLOAT_FORMAT}
    # descriptors
    IonoRangeVertical = FloatDescriptor(
        'IonoRangeVertical', _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere two-way delay error for normal incidence standard deviation. '
                  r'Expressed as a range error. '
                  r':math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: Optional[float]
    IonoRangeRateVertical = FloatDescriptor(
        'IonoRangeRateVertical', _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere two-way delay rate of change error for normal '
                  'incidence standard deviation. Expressed as a range rate error. '
                  r':math:`(\Delta \dot{R}) = (\Delta \dot{T}) \cdot (c/2)`.')  # type: Optional[float]
    IonoRgRgRateCC = FloatDescriptor(
        'IonoRgRgRateCC', _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere range error and range rate error correlation coefficient.')  # type: float
    IonoRangeVertDecorr = SerializableDescriptor(
        'IonoRangeVertDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere range error decorrelation rate.')  # type: Optional[ErrorDecorrFuncType]

    def __init__(
            self,
            IonoRangeVertical: Optional[float] = None,
            IonoRangeRateVertical: Optional[float] = None,
            IonoRgRgRateCC: float = None,
            IonoRangeVertDecorr: Optional[ErrorDecorrFuncType] = None,
            **kwargs):
        """

        Parameters
        ----------
        IonoRangeVertical : float
        IonoRangeRateVertical : float
        IonoRgRgRateCC : float
        IonoRangeVertDecorr : ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.IonoRangeVertical = IonoRangeVertical
        self.IonoRangeRateVertical = IonoRangeRateVertical
        self.IonoRgRgRateCC = IonoRgRgRateCC
        self.IonoRangeVertDecorr = IonoRangeVertDecorr
        super(IonoErrorType, self).__init__(**kwargs)


class ErrorComponentsType(Serializable):
    """Error statistics by components."""
    _fields = ('PosVelErr', 'RadarSensor', 'TropoError', 'IonoError')
    _required = ('PosVelErr', 'RadarSensor')
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the radar platform.')  # type: PosVelErrType
    RadarSensor = SerializableDescriptor(
        'RadarSensor', RadarSensorErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Radar sensor error statistics.')  # type: RadarSensorErrorType
    TropoError = SerializableDescriptor(
        'TropoError', TropoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Troposphere delay error statistics.')  # type: TropoErrorType
    IonoError = SerializableDescriptor(
        'IonoError', IonoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere delay error statistics.')  # type: IonoErrorType

    def __init__(
            self,
            PosVelErr: PosVelErrType = None,
            RadarSensor: RadarSensorErrorType = None,
            TropoError: Optional[TropoErrorType] = None,
            IonoError: Optional[IonoErrorType] = None,
            **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        RadarSensor : RadarSensorErrorType
        TropoError : TropoErrorType
        IonoError : IonoErrorType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr = PosVelErr
        self.RadarSensor = RadarSensor
        self.TropoError = TropoError
        self.IonoError = IonoError
        super(ErrorComponentsType, self).__init__(**kwargs)


class UnmodeledDecorrType(Serializable):
    """
    Unmodeled decorrelation function definition
    """

    _fields = ('Xrow', 'Ycol')
    _required = _fields
    # descriptors
    Xrow = SerializableDescriptor(
        'Xrow', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT)  # type: ErrorDecorrFuncType
    Ycol = SerializableDescriptor(
        'Ycol', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT)  # type: ErrorDecorrFuncType

    def __init__(
            self,
            Xrow: ErrorDecorrFuncType = None,
            Ycol: ErrorDecorrFuncType = None,
            **kwargs):
        """

        Parameters
        ----------
        Xrow : ErrorDecorrFuncType
        Ycol : ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Xrow = Xrow
        self.Ycol = Ycol
        super(UnmodeledDecorrType, self).__init__(**kwargs)


class UnmodeledType(Serializable):
    _fields = ('Xrow', 'Ycol', 'XrowYcol', 'UnmodeledDecorr')
    _required = ('Xrow', 'Ycol', 'XrowYcol')
    _numeric_format = {fld: '0.17G' for fld in ('Xrow', 'Ycol', 'XrowYcol')}
    Xrow = FloatDescriptor(
        'Xrow', _required, strict=DEFAULT_STRICT)  # type: float
    Ycol = FloatDescriptor(
        'Ycol', _required, strict=DEFAULT_STRICT)  # type: float
    XrowYcol = FloatDescriptor(
        'XrowYcol', _required, strict=DEFAULT_STRICT)  # type: float
    UnmodeledDecorr = SerializableDescriptor(
        'UnmodeledDecorr', UnmodeledDecorrType, _required,
        strict=DEFAULT_STRICT)  # type: Optional[UnmodeledDecorrType]

    def __init__(
            self,
            Xrow: float = None,
            Ycol: float = None,
            XrowYcol: float = None,
            UnmodeledDecorr: Optional[UnmodeledDecorrType] = None,
            **kwargs):
        """

        Parameters
        ----------
        Xrow : float
        Ycol : float
        XrowYcol : float
        UnmodeledDecorr : None|UnmodeledDecorrType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Xrow = Xrow
        self.Ycol = Ycol
        self.XrowYcol = XrowYcol
        self.UnmodeledDecorr = UnmodeledDecorr
        super(UnmodeledType, self).__init__(**kwargs)


class ErrorStatisticsType(Serializable):
    """Parameters used to compute error statistics within the SICD sensor model."""
    _fields = ('CompositeSCP', 'Components', 'Unmodeled', 'AdditionalParms')
    _required = ()
    _collections_tags = {'AdditionalParms': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    CompositeSCP = SerializableDescriptor(
        'CompositeSCP', CompositeSCPErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Composite error statistics for the scene center point. '
                  '*Slant plane range (Rg)* and *azimuth (Az)* error statistics. '
                  'Slant plane defined at '
                  '*Scene Center Point, Center of Azimuth (SCP COA)*.')  # type: Optional[CompositeSCPErrorType]
    Components = SerializableDescriptor(
        'Components', ErrorComponentsType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics by components.')  # type: Optional[ErrorComponentsType]
    Unmodeled = SerializableDescriptor(
        'Unmodeled', UnmodeledType, _required, strict=DEFAULT_STRICT)  # type: Optional[UnmodeledType]

    AdditionalParms = ParametersDescriptor(
        'AdditionalParms', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Any additional parameters.')  # type: Optional[ParametersCollection]

    def __init__(
            self,
            CompositeSCP: Optional[CompositeSCPErrorType] = None,
            Components: Optional[ErrorComponentsType] = None,
            Unmodeled: Optional[UnmodeledType] = None,
            AdditionalParms: Union[None, ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        CompositeSCP : None|CompositeSCPErrorType
        Components : None|ErrorComponentsType
        Unmodeled : None|UnmodeledType
        AdditionalParms : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CompositeSCP = CompositeSCP
        self.Components = Components
        self.Unmodeled = Unmodeled
        self.AdditionalParms = AdditionalParms
        super(ErrorStatisticsType, self).__init__(**kwargs)

    def version_required(self):
        """
        What SICD version is required?

        Returns
        -------
        Tuple[int, int, int]
        """

        return (1, 1, 0) if self.Unmodeled is None else (1, 3, 0)



"""
The image subheader definitions.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import struct
from collections import OrderedDict
from typing import Union

import numpy

from .base import NITFElement, NITFLoop, UserHeaderType, _IntegerDescriptor,\
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor, _parse_str
from .security import NITFSecurityTags, NITFSecurityTags0


logger = logging.getLogger(__name__)


######
# General components

class ImageBand(NITFElement):
    """
    Single image band, part of the image bands collection
    """
    _ordering = ('IREPBAND', 'ISUBCAT', 'IFC', 'IMFLT', 'LUTD')
    _lengths = {'IREPBAND': 2, 'ISUBCAT': 6, 'IFC': 1, 'IMFLT': 3}
    IREPBAND = _StringDescriptor(
        'IREPBAND', True, 2, default_value='',
        docstring='Representation. This field shall contain a valid indicator of the processing '
                  'required to display the nth band of the image with regard to the general image type '
                  'as recorded in the `IREP` field. The significance of each band in the image can be '
                  'derived from the combination of the `ICAT`, and `ISUBCAT` fields. Valid values of '
                  'the `IREPBAND` field depend on the value of '
                  'the `IREP` field.')  # type: str
    ISUBCAT = _StringDescriptor(
        'ISUBCAT', True, 6, default_value='',
        docstring='Subcategory. The purpose of this field is to provide the significance of the band '
                  'of the image with regard to the specific category (`ICAT` field) '
                  'of the overall image.')  # type: str
    IFC = _StringEnumDescriptor(
        'IFC', True, 1, {'N', }, default_value='N',
        docstring=' Image Filter Condition.')  # type: str
    IMFLT = _StringDescriptor(
        'IMFLT', True, 3, default_value='',
        docstring='Standard Image Filter Code. This field is reserved '
                  'for future use.')  # type: str

    def __init__(self, **kwargs):
        self._LUTD = None
        super(ImageBand, self).__init__(**kwargs)

    @classmethod
    def minimum_length(cls):
        return 13

    @property
    def LUTD(self):
        """
        The Look-up Table (LUT) data.

        Returns
        -------
        None|numpy.ndarray
        """

        return self._LUTD

    @LUTD.setter
    def LUTD(self, value):
        if value is None:
            self._LUTD = None
            return

        if not isinstance(value, numpy.ndarray):
            raise TypeError('LUTD must be a numpy array')
        if value.dtype.name != 'uint8':
            raise ValueError('LUTD must be a numpy array of dtype uint8, got {}'.format(value.dtype.name))
        if value.ndim != 2:
            raise ValueError('LUTD must be a two-dimensional array')
        if value.shape[0] > 4:
            raise ValueError(
                'The number of LUTD bands (axis 0) must be 4 or fewer. '
                'Got LUTD shape {}'.format(value.shape))
        if value.shape[1] > 65536:
            raise ValueError(
                'The number of LUTD elemnts (axis 1) must be 65536 or fewer. '
                'Got LUTD shape {}'.format(value.shape))
        self._LUTD = value

    @property
    def NLUTS(self):
        """
        Number of LUTS for the Image Band. This field shall contain the number
        of LUTs associated with the nth band of the image. LUTs are allowed
        only if the value of the `PVTYPE` field is :code:`INT` or :code:`B`.

        Returns
        -------
        int
        """

        return 0 if self._LUTD is None else self._LUTD.shape[0]

    @property
    def NELUTS(self):
        """
        Number of LUT Entries for the Image Band. This field shall contain
        the number of entries in each of the LUTs for the nth image band.

        Returns
        -------
        int
        """

        return 0 if self._LUTD is None else self._LUTD.shape[1]

    def _get_attribute_bytes(self, attribute):
        if attribute == 'LUTD':
            if self.NLUTS == 0:
                out = b'0'
            else:
                out = '{0:d}{1:05d}'.format(self.NLUTS, self.NELUTS).encode() + \
                      struct.pack('{}B'.format(self.NLUTS * self.NELUTS), *self.LUTD.flatten())
            return out
        else:
            return super(ImageBand, self)._get_attribute_bytes(attribute)

    def _get_attribute_length(self, attribute):
        if attribute == 'LUTD':
            nluts = self.NLUTS
            if nluts == 0:
                return 1
            else:
                neluts = self.NELUTS
                return 6 + nluts * neluts
        else:
            return super(ImageBand, self)._get_attribute_length(attribute)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'LUTD':
            loc = start
            nluts = int(value[loc:loc + 1])
            loc += 1
            if nluts == 0:
                fields['LUTD'] = None
            else:
                neluts = int(value[loc:loc + 5])
                loc += 5
                siz = nluts * neluts
                lutd = numpy.array(
                    struct.unpack('{}B'.format(siz), value[loc:loc + siz]), dtype=numpy.uint8).reshape(
                    (nluts, neluts))
                fields['LUTD'] = lutd
                loc += siz
            return loc
        return super(ImageBand, cls)._parse_attribute(fields, attribute, value, start)


class ImageBands(NITFLoop):
    _child_class = ImageBand
    _count_size = 1
    NBANDS_LEN = 1
    XBANDS_LEN = 5

    @NITFLoop.values.setter
    def values(self, value):
        NITFLoop.values.fset(self, value)
        if value and len(value) > 9:
            self._count_size = self.NBANDS_LEN + self.XBANDS_LEN
        else:
            self._count_size = self.NBANDS_LEN

    @classmethod
    def _parse_count(cls, value, start):
        count, loc = super()._parse_count(value, start)
        if count == 0:
            # (only) if there are more than 9, a longer field is used
            count = int(value[loc:loc + cls.XBANDS_LEN])
            loc += cls.XBANDS_LEN
        return count, loc

    @classmethod
    def minimum_length(cls):
        return cls.NBANDS_LEN + cls._child_class.minimum_length()


class ImageComment(NITFElement):
    _ordering = ('COMMENT', )
    _lengths = {'COMMENT': 80}
    COMMENT = _StringDescriptor('COMMENT', True, 80, default_value='', docstring='The image comment')


class ImageComments(NITFLoop):
    _child_class = ImageComment
    _count_size = 1


########
# Masked image header - this is a binary structure

class MaskSubheader(NITFElement):
    _ordering = (
        'IMDATOFF', 'BMRLNTH', 'TMRLNTH', 'TPXCDLNTH', 'TPXCD', 'BMR', 'TMR')
    _lengths = {
        'IMDATOFF': 4, 'BMRLNTH': 2, 'TMRLNTH': 2, 'TPXCDLNTH': 2}
    _binary_format = {
        'IMDATOFF': '>I', 'BMRLNTH': '>H', 'TMRLNTH': '>H', 'TPXCDLNTH': '>H'}
    # descriptors
    IMDATOFF = _IntegerDescriptor(
        'IMDATOFF', True, 10,
        docstring='Blocked image data offset. This is the size of the masked subheader '
                  'in bytes.')  # type: int
    BMRLNTH = _IntegerDescriptor(
        'BMRLNTH', True, 5,
        docstring='Block mask record length')  # type: int
    TMRLNTH = _IntegerDescriptor(
        'TMRLNTH', True, 5,
        docstring='Transparent Pixel Mask Record Length')  # type: int
    TPXCDLNTH = _IntegerDescriptor(
        'TPXCDLNTH', True, 5,
        docstring='Transparent Output Pixel Code Length in bits.')  # type: int

    def __init__(self, band_depth=1, blocks=1, **kwargs):
        self._band_depth = band_depth
        self._blocks = blocks
        self._TPXCD = None
        self._BMR = None
        self._TMR = None
        super(MaskSubheader, self).__init__(**kwargs)

    @property
    def band_depth(self):
        """
        int: The number of band elements. Will only be > 1 if band-sequential format.
        """

        return self._band_depth

    @property
    def blocks(self):
        """
        int: The number of blocks.
        """

        return self._blocks

    @property
    def TPXCD(self):
        """
        bytes: The Transparent output pixel code.
        """

        return self._TPXCD

    @TPXCD.setter
    def TPXCD(self, value):
        if self.TPXCDLNTH == 0:
            self._TPXCD = None
            return

        if not isinstance(value, bytes):
            raise TypeError('TPXCD must be of type bytes.')
        expected_length = self._get_attribute_length('TPXCD')
        if len(value) != expected_length:
            raise ValueError(
                'Provided TPXCD data is required to be of length {}, '
                'but got length {}'.format(expected_length, len(value)))
        self._TPXCD = value

    @property
    def BMR(self):
        # type: () -> Union[None, numpy.ndarray]
        """
        None|numpy.ndarray: The block mask records array. This will be None if
        and only if `BMRLNTH=0`. Each entry records the offset in bytes from
        the beginning of the blocked image data to the first byte of the respective
        block. If the block is not recorded/transmitted (i.e. present), then the
        offset value is defaulted to `0xFFFFFFFF`.
        """

        return self._BMR

    @BMR.setter
    def BMR(self, value):
        if value is None:
            if self.BMRLNTH != 0:
                raise ValueError('BMR array is None, but BMRLNTH={}'.format(self.BMRLNTH))
            self._BMR = None
        else:
            if self.BMRLNTH != 4:
                raise ValueError('BMR array is provided, but BMRLNTH={}'.format(self.BMRLNTH))
            if not isinstance(value, numpy.ndarray):
                value = numpy.array(value, dtype='uint32')
            if value.shape != (self.band_depth, self.blocks):
                raise ValueError(
                    'BMR array is of shape {}, and must be of '
                    'shape {}'.format(value.shape, (self.band_depth, self.blocks)))
            if value.dtype.name != 'uint32':
                raise ValueError(
                    'BMR array has dtype {}, and must be of '
                    'dtype uint32'.format(value.dtype.name))
            self._BMR = value

    @property
    def TMR(self):
        # type: () -> Union[None, numpy.ndarray]
        """
        None|numpy.ndarray: The transparent mask records array. This will be None if
        and only if `TMRLNTH=0`. Each entry records the offset in bytes from
        the beginning of the blocked image data to the first byte of the respective
        block (if this block contains pad pixels), or the default value `0xFFFFFFFF`
        to indicate that this block does not contain pad pixels.
        """

        return self._TMR

    @TMR.setter
    def TMR(self, value):
        if value is None:
            if self.TMRLNTH != 0:
                raise ValueError('TMR array is None, but TMRLNTH={}'.format(self.TMRLNTH))
            self._TMR = None
        else:
            if self.TMRLNTH != 4:
                raise ValueError('TMR array is provided, but TMRLNTH={}'.format(self.TMRLNTH))
            if not isinstance(value, numpy.ndarray):
                value = numpy.array(value, dtype='uint32')
            if value.shape != (self.band_depth, self.blocks):
                raise ValueError(
                    'TMR array is of shape {}, and must be of '
                    'shape {}'.format(value.shape, (self.band_depth, self.blocks)))
            if value.dtype.name != 'uint32':
                raise ValueError(
                    'TMR array has dtype {}, and must be of '
                    'dtype uint32'.format(value.dtype.name))
            self._TMR = value

    @staticmethod
    def define_tpxcd_length(tpxcdlnth):
        """
        Gets the appropriate length for the TPXCD data.

        Parameters
        ----------
        tpxcdlnth : int
            The TPXCDLNTH value.

        Returns
        -------
        int
        """

        missing = (tpxcdlnth % 8)
        if missing == 0:
            return int(tpxcdlnth/8)
        else:
            return int((tpxcdlnth + (8 - missing))/8)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'BMR':
            if fields['BMRLNTH'] == 0:
                fields['BMR'] = None
                return start
            else:
                count = fields['band_depth']*fields['blocks']
                end = start+4*count
                array = numpy.array(struct.unpack('>{}I'.format(count), value[start:end]), dtype='uint32')
                fields['BMR'] = numpy.resize(array, (fields['band_depth'], fields['blocks']))
                return end
        elif attribute == 'TMR':
            if fields['TMRLNTH'] == 0:
                fields['TMR'] = None
                return start
            else:
                count = fields['band_depth']*fields['blocks']
                end = start+4*count
                array = numpy.array(struct.unpack('>{}I'.format(count), value[start:end]), dtype='uint32')
                fields['TMR'] = numpy.resize(array, (fields['band_depth'], fields['blocks']))
                return end
        elif attribute == 'TPXCD':
            length = cls.define_tpxcd_length(fields['TPXCDLNTH'])
            if length == 0:
                fields['TPXCD'] = None
                return start
            else:
                end = start + length
                fields['TPXCD'] = value[start:end]
                return end
        else:
            return super(MaskSubheader, cls)._parse_attribute(fields, attribute, value, start)

    def _get_attribute_length(self, fld):
        if fld in ['BMR', 'TMR']:
            value = getattr(self, fld)
            if value is None:
                return 0
            else:
                return value.size*4
        elif fld == 'TPXCD':
            return self.define_tpxcd_length(self.TPXCDLNTH)
        else:
            return super(MaskSubheader, self)._get_attribute_length(fld)

    def _get_attribute_bytes(self, fld):
        if fld in ['BMR', 'TMR']:
            value = getattr(self, fld)
            if value is None:
                return b''
            else:
                return struct.pack('>{}I'.format(value.size), *numpy.reshape(value, (-1,)))
        elif fld == 'TPXCD':
            if self._TPXCD is None:
                return b''
            return self._TPXCD
        else:
            return super(MaskSubheader, self)._get_attribute_bytes(fld)

    @classmethod
    def from_bytes(cls, value, start, band_depth=1, blocks=1):
        fields = {
            'band_depth': band_depth, 'blocks': blocks}
        loc = start
        for fld in cls._ordering:
            loc = cls._parse_attribute(fields, fld, value, loc)
        out = cls(**fields)

        input_length = len(value)-start
        out_length = out.get_bytes_length()
        if input_length != out_length:
            logger.error(
                'The MaskSubheader object is being serialized from a bytes buffer of length {},\n\t'
                'but would serialize to a bytes object of length {}.\n\t'
                'This is likely a result of faulty serialization,\n\t '
                'and represents an error.'.format(input_length, out_length))
        return out

    def to_json(self):
        out = OrderedDict([('band_depth', self.band_depth), ('blocks', self.blocks)])
        for fld in self._ordering:
            value = getattr(self, fld)
            if value is None:
                continue
            if fld in ['BMR', 'TMR']:
                out[fld] = value.tolist()
            else:
                out[fld] = value
        return out


#########
# NITF 2.1 version

class ImageSegmentHeader(NITFElement):
    """
    The image segment header - see standards document Joint BIIF Profile (JBP) for more
    information.
    """

    _ordering = (
        'IM', 'IID1', 'IDATIM', 'TGTID',
        'IID2', 'Security', 'ENCRYP', 'ISORCE',
        'NROWS', 'NCOLS', 'PVTYPE', 'IREP',
        'ICAT', 'ABPP', 'PJUST', 'ICORDS',
        'IGEOLO', 'Comments', 'IC', 'COMRAT', 'Bands',
        'ISYNC', 'IMODE', 'NBPR', 'NBPC', 'NPPBH',
        'NPPBV', 'NBPP', 'IDLVL', 'IALVL',
        'ILOC', 'IMAG', 'UserHeader', 'ExtendedHeader')
    _lengths = {
        'IM': 2, 'IID1': 10, 'IDATIM': 14, 'TGTID': 17,
        'IID2': 80, 'ENCRYP': 1, 'ISORCE': 42,
        'NROWS': 8, 'NCOLS': 8, 'PVTYPE': 3, 'IREP': 8,
        'ICAT': 8, 'ABPP': 2, 'PJUST': 1, 'ICORDS': 1,
        'IGEOLO': 60, 'IC': 2, 'COMRAT': 4, 'ISYNC': 1, 'IMODE': 1,
        'NBPR': 4, 'NBPC': 4, 'NPPBH': 4, 'NPPBV': 4,
        'NBPP': 2, 'IDLVL': 3, 'IALVL': 3, 'ILOC': 10,
        'IMAG': 4, 'UDIDL': 5, 'IXSHDL': 5}
    # Descriptors
    IM = _StringEnumDescriptor(
        'IM', True, 2, {'IM', }, default_value='IM',
        docstring='File part type.')  # type: str
    IID1 = _StringDescriptor(
        'IID1', True, 10, default_value='',
        docstring='Image Identifier 1. This field shall contain a valid alphanumeric identification code '
                  'associated with the image. The valid codes are determined by '
                  'the application.')  # type: str
    IDATIM = _StringDescriptor(
        'IDATIM', True, 14, default_value='',
        docstring='Image Date and Time. This field shall contain the time (UTC) of the image '
                  'acquisition in the format :code:`YYYYMMDDhhmmss`.')  # type: str
    TGTID = _StringDescriptor(
        'TGTID', True, 17, default_value='',
        docstring='Target Identifier. This field shall contain the identification of the primary target '
                  'in the format, :code:`BBBBBBBBBBOOOOOCC`, consisting of ten characters of Basic Encyclopedia '
                  '`(BE)` identifier, followed by five characters of facility OSUFFIX, followed by the two '
                  'character country code as specified in FIPS PUB 10-4.')  # type: str
    IID2 = _StringDescriptor(
        'IID2', True, 80, default_value='',
        docstring='Image Identifier 2. This field can contain the identification of additional '
                  'information about the image.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags, default_args={},
        docstring='The image security tags.')  # type: NITFSecurityTags
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    ISORCE = _StringDescriptor(
        'ISORCE', True, 42, default_value='',
        docstring='Image Source. This field shall contain a description of the source of the image. '
                  'If the source of the data is classified, then the description shall be preceded by '
                  'the classification, including codeword(s).')  # type: str
    NROWS = _IntegerDescriptor(
        'NROWS', True, 8, default_value=0,
        docstring='Number of Significant Rows in Image. This field shall contain the total number of rows '
                  'of significant pixels in the image. When the product of the values of the `NPPBV` field '
                  'and the `NBPC` field is greater than the value of the `NROWS` field '
                  r'(:math:`NPPBV \cdot NBPC > NROWS`), the rows indexed with the value of the `NROWS` field '
                  r'to (:math:`NPPBV\cdot NBPC - 1`) shall contain fill data. NOTE: Only the rows indexed '
                  '0 to the value of the `NROWS` field minus 1 of the image contain significant data. '
                  'The pixel fill values are determined by the application.')  # type: int
    NCOLS = _IntegerDescriptor(
        'NCOLS', True, 8, default_value=0,
        docstring='Number of Significant Columns in Image. This field shall contain the total number of '
                  'columns of significant pixels in the image. When the product of the values of the `NPPBH` '
                  'field and the `NBPR` field is greater than the `NCOLS` field '
                  r'(:math:`NPPBH\cdot NBPR > NCOLS`), the columns indexed with the value of the `NCOLS` field '
                  r'to (:math:`NPPBH\cdot NBPR - 1`) shall contain fill data. NOTE: Only the columns '
                  'indexed 0 to the value of the `NCOLS` field minus 1 of the image contain significant data. '
                  'The pixel fill values are determined by the application.')  # type: int
    PVTYPE = _StringEnumDescriptor(
        'PVTYPE', True, 3, {'INT', 'B', 'SI', 'R', 'C'},
        docstring='Pixel Value Type. This field shall contain an indicator of the type of computer representation '
                  'used for the value for each pixel for each band in the image. ')  # type: str
    IREP = _StringEnumDescriptor(
        'IREP', True, 8,
        {'MONO', 'RGB', 'RGB/LUT', 'MULTI', 'NODISPLY', 'NVECTOR', 'POLAR', 'VPH', 'YCbCr601'},
        default_value='NODISPLY',
        docstring='Image Representation. This field shall contain a valid indicator of the processing required '
                  'in order to display an image.')  # type: str
    ICAT = _StringDescriptor(
        'ICAT', True, 8, default_value='SAR',
        docstring='Image Category. This field shall contain a valid indicator of the specific category of image, '
                  'raster or grid data. The specific category of an IS reveals its intended use or the nature '
                  'of its collector.')  # type: str
    ABPP = _IntegerDescriptor(
        'ABPP', True, 2,
        docstring='Actual Bits-Per-Pixel Per Band. This field shall contain the number of "significant bits" for '
                  'the value in each band of each pixel without compression. Even when the image is compressed, '
                  '`ABPP` contains the number of significant bits per pixel that were present in the image '
                  'before compression. This field shall be less than or equal to Number of Bits Per Pixel '
                  '(field `NBPP`). The number of adjacent bits within each `NBPP` is '
                  'used to represent the value.')  # type: int
    PJUST = _StringEnumDescriptor(
        'PJUST', True, 1, {'L', 'R'}, default_value='R',
        docstring='Pixel Justification. When `ABPP` is not equal to `NBPP`, this field indicates whether the '
                  'significant bits are left justified (:code:`L`) or right '
                  'justified (:code:`R`).')  # type: str
    ICORDS = _StringEnumDescriptor(
        'ICORDS', True, 1, {'', 'U', 'G', 'N', 'S', 'D'}, default_value='G',
        docstring='Image Coordinate Representation. This field shall contain a valid code indicating the type '
                  'of coordinate representation used for providing an approximate location of the image in the '
                  'Image Geographic Location field (`IGEOLO`).')  # type: str
    Comments = _NITFElementDescriptor(
        'Comments', True, ImageComments, default_args={},
        docstring='The image comments.')  # type: ImageComments
    Bands = _NITFElementDescriptor(
        'Bands', True, ImageBands, default_args={},
        docstring='The image bands.')  # type: ImageBands
    ISYNC = _IntegerDescriptor(
        'ISYNC', True, 1, default_value=0,
        docstring='Image Sync code. This field is reserved for future use. ')  # type: int
    IMODE = _StringEnumDescriptor(
        'IMODE', True, 1, {'B', 'P', 'R', 'S'}, default_value='P',
        docstring='Image Mode. This field shall indicate how the Image Pixels are '
                  'stored in the NITF file.')  # type: str
    NBPR = _IntegerDescriptor(
        'NBPR', True, 4, default_value=1,
        docstring='Number of Blocks Per Row. This field shall contain the number of image blocks in a row of '
                  'blocks (paragraph 5.4.2.2) in the horizontal direction. If the image consists of only a '
                  'single block, this field shall contain the value one.')  # type: int
    NBPC = _IntegerDescriptor(
        'NBPC', True, 4, default_value=1,
        docstring='Number of Blocks Per Column. This field shall contain the number of image blocks in a column '
                  'of blocks (paragraph 5.4.2.2) in the vertical direction. If the image consists of only a '
                  'single block, this field shall contain the value one.')  # type: int
    NPPBH = _IntegerDescriptor(
        'NPPBH', True, 4, default_value=0,
        docstring='Number of Pixels Per Block Horizontal. This field shall contain the number of pixels horizontally '
                  'in each block of the image. It shall be the case that the product of the values of the `NBPR` '
                  'field and the `NPPBH` field is greater than or equal to the value of the `NCOLS` field '
                  r'(:math:`NBPR\cdot NPPBH \geq NCOLS`). When NBPR is :code:`1`, setting the `NPPBH` '
                  'value to :code:`0` designates that the number of pixels horizontally is specified by the '
                  'value in NCOLS.')  # type: int
    NPPBV = _IntegerDescriptor(
        'NPPBV', True, 4, default_value=0,
        docstring='Number of Pixels Per Block Vertical. This field shall contain the number of pixels vertically '
                  'in each block of the image. It shall be the case that the product of the values of the `NBPC` '
                  'field and the `NPPBV` field is greater than or equal to the value of the `NROWS` field '
                  r'(:math:`NBPC\cdot NPPBV \geq NROWS`). When `NBPC` is :code:`1`, setting the `NPPBV` value '
                  r'to :code:`0` designates that the number of pixels vertically is specified by '
                  r'the value in `NROWS`.')  # type: int
    NBPP = _IntegerDescriptor(
        'NBPP', True, 2, default_value=0,
        docstring='Number of Bits Per Pixel Per Band.')  # type: int
    IDLVL = _IntegerDescriptor(
        'IDLVL', True, 3, default_value=0,
        docstring='Image Display Level. This field shall contain a valid value that indicates the display level of '
                  'the image relative to other displayed file components in a composite display. The valid values '
                  'are :code:`1-999`. The display level of each displayable segment (image or graphic) within a file '
                  'shall be unique.')  # type: int
    IALVL = _IntegerDescriptor(
        'IALVL', True, 3, default_value=0,
        docstring='Attachment Level. This field shall contain a valid value that indicates the attachment '
                  'level of the image.')  # type: int
    ILOC = _StringDescriptor(
        'ILOC', True, 10, default_value='',
        docstring='Image Location. The image location is the location of the first pixel of the first line of the '
                  'image. This field shall contain the image location offset from the `ILOC` or `SLOC` value '
                  'of the segment to which the image is attached or from the origin of the CCS when the image '
                  'is unattached (`IALVL` contains :code:`0`). A row or column value of :code:`0` indicates no offset. '
                  'Positive row and column values indicate offsets down and to the right while negative row and '
                  'column values indicate offsets up and to the left.')  # type: str
    IMAG = _StringDescriptor(
        'IMAG', True, 4, default_value='1.0',
        docstring='Image Magnification. This field shall contain the magnification (or reduction) factor of the '
                  'image relative to the original source image. Decimal values are used to indicate magnification, '
                  'and decimal fraction values indicate reduction. For example, :code:`2.30` indicates the original '
                  'image has been magnified by a factor of :code:`2.30`, while :code:`0.5` indicates '
                  'the original image has been reduced by a factor of 2.')  # type: str
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType
    ExtendedHeader = _NITFElementDescriptor(
        'ExtendedHeader', True, UserHeaderType, default_args={},
        docstring='Extended subheader - TRE list.')  # type: UserHeaderType

    def __init__(self, **kwargs):
        self._IC = None
        self._COMRAT = None
        self._IGEOLO = None
        self._mask_subheader = None
        super(ImageSegmentHeader, self).__init__(**kwargs)

    @property
    def is_masked(self):
        """
        bool: Does this image segment contain a mask?
        """

        return self.IC in ['NM', 'M1', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8']

    @property
    def is_compressed(self):
        """
        bool: Is this image segment compressed?
        """

        return self.IC not in ['NC', 'NM']

    @property
    def IC(self):
        """
        str: Image Compression. This field shall contain a valid code indicating
        the form of compression used in representing the image data.

        Valid values for this field are, :code:`C1` to represent bi-level, :code:`C3`
        to represent JPEG, :code:`C4` to represent Vector Quantization, :code:`C5`
        to represent lossless JPEG, :code:`I1` to represent down sampled JPEG,
        and :code:`NC` to represent the image is not compressed. Also valid are
        :code:`M1, M3, M4`, and :code:`M5` for compressed images, and :code:`NM`
        for uncompressed images indicating an image that contains a block
        mask and/or a pad pixel mask. :code:`C6` and :code:`M6` are reserved values
        that will represent a future correlated multicomponent compression
        algorithm. :code:`C7` and :code:`M7` are reserved values that will represent
        a future complex SAR compression. :code:`C8` and :code:`M8` are the values
        for ISO standard compression JPEG 2000.

        The format of a mask image is identical to the format of its corresponding non-masked image
        except for the presence of an Image Data Mask at the beginning of
        the image data area. The format of the Image Data Mask is described
        in paragraph 5.4.3.2 and is shown in table A-3(A). The definitions
        of the compression schemes associated with codes :code:`C1/M1, C3/M3, C4/M4, C5/M5`
        are given, respectively, in ITU- T T.4, AMD2, MIL-STD-188-198A,
        MIL-STD- 188-199, and NGA N0106-97. :code:`C1` is found in ITU- T T.4 AMD2,
        :code:`C3` is found in MIL-STD-188-198A, :code:`C4` is found in MIL-STD-188-199,
        and :code:`C5` and :code:`I1` are found in NGA N0106-97. (NOTE: :code:`C2` (ARIDPCM) is not
        valid in NITF 2.1.) The definition of the compression scheme associated
        with codes :code:`C8/M8` is found in ISO/IEC 15444- 1:2000 (with amendments 1 and 2).
        """

        return self._IC

    @IC.setter
    def IC(self, value):
        value = _parse_str(value, 2, 'NC', 'IC', self)
        if value not in {
                'NC', 'NM', 'C0', 'C1', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'I1',
                'M1', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8'}:
            raise ValueError('IC got invalid value {}'.format(value))
        self._IC = value
        if value in ('NC', 'NM'):
            self._COMRAT = None
        elif self._COMRAT is not None:
            self._COMRAT = '\x20'*4

    @property
    def COMRAT(self):
        """
        None|str: Compression Rate Code. If the IC field contains one of
        :code:`C1, C3, C4, C5, C8, M1, M3, M4, M5, M8, I1`, this field shall be contain
        a code indicating the compression rate for the image.

        If `IC` is :code:`NC` or :code:`NM`, then this will be set to :code:`None`.
        """

        return self._COMRAT

    @COMRAT.setter
    def COMRAT(self, value):
        value = _parse_str(value, 4, None, 'COMRAT', self)
        if value is None and self.IC not in ('NC', 'NM'):
            value = '\x20'*4
            logger.error(
                'COMRAT value is None, but IC is not in {"NC", "NM"}.\n\t'
                'This must be resolved.')
        if value is not None and self.IC in ('NC', 'NM'):
            value = None
            logger.error(
                'COMRAT value is something other than None, but IC in {"NC", "NM"}.\n\t'
                'This is invalid, and COMRAT is being set to None.')
        self._COMRAT = value

    @property
    def IGEOLO(self):
        """
        None|str: Image Geographic Location. This field, when present, shall contain
        an approximate geographic location which is not intended for analytical purposes
        (e.g., targeting, mensuration, distance calculation); it is intended to support
        general user appreciation for the image location (e.g., cataloguing). The
        representation of the image corner locations is specified in the `ICORDS` field.
        The locations of the four corners of the (significant) image data shall be given
        in image coordinate order: (0,0), (0, MaxCol), (MaxRow, MaxCol), (MaxRow, 0).
        MaxCol and MaxRow shall be determined from the values contained, respectively,
        in the `NCOLS` field and the `NROWS` field.
        """

        return self._IGEOLO

    @IGEOLO.setter
    def IGEOLO(self, value):
        value = _parse_str(value, 60, None, 'IGEOLO', self)
        if value is None and self.ICORDS.strip() != '':
            value = '\x20'*60
        if value is not None and self.ICORDS.strip() == '':
            value = None
        self._IGEOLO = value

    @property
    def mask_subheader(self):
        # type: () -> Union[None, MaskSubheader]
        """
        None|MaskSubheader: The mask subheader, if it has been appended.
        """

        return self._mask_subheader

    @mask_subheader.setter
    def mask_subheader(self, value):
        if value is None:
            self._mask_subheader = None
            return
        if not isinstance(value, MaskSubheader):
            raise ValueError(
                'mask_subheader is expected to be an instance of MaskSubheader. '
                'Got type {}'.format(type(value)))
        if self.IC not in ['NM', 'M1', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8']:
            raise ValueError(
                'IC={}, which does not indicate the presence of a mask '
                'subheader'.format(self.IC))
        self._mask_subheader = value

    def _get_attribute_length(self, fld):
        if fld in ['COMRAT', 'IGEOLO']:
            if getattr(self, '_'+fld) is None:
                return 0
            else:
                return self._lengths[fld]
        else:
            return super(ImageSegmentHeader, self)._get_attribute_length(fld)

    @classmethod
    def minimum_length(cls):
        # COMRAT and IGEOLO may not be there
        return super(ImageSegmentHeader, cls).minimum_length() - 64

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'IC':
            val = value[start:start+2].decode('utf-8')
            fields['IC'] = val
            if val in ('NC', 'NM'):
                fields['COMRAT'] = None
            return start+2
        elif attribute == 'ICORDS':
            fields['ICORDS'] = value[start:start+1]
            if fields['ICORDS'] == b' ':
                fields['IGEOLO'] = None
            return start+1
        else:
            return super(ImageSegmentHeader, cls)._parse_attribute(fields, attribute, value, start)

    def get_uncompressed_block_size(self) -> int:
        """
        Gets the size of an uncompressed block.

        Note that if `IMODE == 'S'`, then each block consists of a single band.
        Otherwise, a block consists of all bands.

        Returns
        -------
        int
        """

        nppbv = self.NROWS if self.NPPBV == 0 else self.NPPBV
        nppbh = self.NCOLS if self.NPPBH == 0 else self.NPPBH
        if self.IMODE == 'S':
            return int(nppbh*nppbv*self.NBPP/8)
        else:
            return int(nppbh*nppbv*len(self.Bands)*self.NBPP/8)

    def get_full_uncompressed_image_size(self) -> int:
        """
        Gets the full size in bytes of the uncompressed image including any padding
        in the blocks.

        Returns
        -------
        int
        """

        total_blocks = self.NBPR*self.NBPC
        if self.IMODE == 'S':
            total_blocks *= len(self.Bands)
        return total_blocks*self.get_uncompressed_block_size()

    def get_clevel(self) -> int:
        """
        Gets the CLEVEL value for this image segment.

        Returns
        -------
        int
        """

        dim_size = max(self.NROWS, self.NCOLS)
        if dim_size <= 2048:
            return 3
        elif dim_size <= 8192:
            return 5
        elif dim_size <= 65536:
            return 6
        else:
            return 7


#########
# NITF 2.0 version

class ImageSegmentHeader0(NITFElement):
    """
    The image segment header for NITF version 2.0 - see standards document
    MIL-STD-2500A for more information.
    """

    _ordering = (
        'IM', 'IID', 'IDATIM', 'TGTID',
        'ITITLE', 'Security', 'ENCRYP', 'ISORCE',
        'NROWS', 'NCOLS', 'PVTYPE', 'IREP',
        'ICAT', 'ABPP', 'PJUST', 'ICORDS',
        'IGEOLO', 'Comments', 'IC', 'COMRAT', 'Bands',
        'ISYNC', 'IMODE', 'NBPR', 'NBPC', 'NPPBH',
        'NPPBV', 'NBPP', 'IDLVL', 'IALVL',
        'ILOC', 'IMAG', 'UserHeader', 'ExtendedHeader')
    _lengths = {
        'IM': 2, 'IID': 10, 'IDATIM': 14, 'TGTID': 17,
        'ITITLE': 80, 'ENCRYP': 1, 'ISORCE': 42,
        'NROWS': 8, 'NCOLS': 8, 'PVTYPE': 3, 'IREP': 8,
        'ICAT': 8, 'ABPP': 2, 'PJUST': 1, 'ICORDS': 1,
        'IGEOLO': 60, 'IC': 2, 'COMRAT': 4, 'ISYNC': 1, 'IMODE': 1,
        'NBPR': 4, 'NBPC': 4, 'NPPBH': 4, 'NPPBV': 4,
        'NBPP': 2, 'IDLVL': 3, 'IALVL': 3, 'ILOC': 10,
        'IMAG': 4, 'UDIDL': 5, 'IXSHDL': 5}
    # Descriptors
    IM = _StringEnumDescriptor(
        'IM', True, 2, {'IM', }, default_value='IM',
        docstring='File part type.')  # type: str
    IID = _StringDescriptor(
        'IID', True, 10, default_value='',
        docstring='Image Identifier 1. This field shall contain a valid alphanumeric identification code '
                  'associated with the image. The valid codes are determined by '
                  'the application.')  # type: str
    IDATIM = _StringDescriptor(
        'IDATIM', True, 14, default_value='',
        docstring='Image Date and Time. This field shall contain the time (UTC) of the image '
                  'acquisition in the format :code:`YYYYMMDDhhmmss`.')  # type: str
    TGTID = _StringDescriptor(
        'TGTID', True, 17, default_value='',
        docstring='Target Identifier. This field shall contain the identification of the primary target '
                  'in the format, :code:`BBBBBBBBBBOOOOOCC`, consisting of ten characters of Basic Encyclopedia '
                  '`(BE)` identifier, followed by five characters of facility OSUFFIX, followed by the two '
                  'character country code as specified in FIPS PUB 10-4.')  # type: str
    ITITLE = _StringDescriptor(
        'ITITLE', True, 80, default_value='',
        docstring='Image Identifier 2. This field can contain the identification of additional '
                  'information about the image.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags0, default_args={},
        docstring='The image security tags.')  # type: NITFSecurityTags0
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    ISORCE = _StringDescriptor(
        'ISORCE', True, 42, default_value='',
        docstring='Image Source. This field shall contain a description of the source of the image. '
                  'If the source of the data is classified, then the description shall be preceded by '
                  'the classification, including codeword(s).')  # type: str
    NROWS = _IntegerDescriptor(
        'NROWS', True, 8, default_value=0,
        docstring='Number of Significant Rows in Image. This field shall contain the total number of rows '
                  'of significant pixels in the image. When the product of the values of the `NPPBV` field '
                  'and the `NBPC` field is greater than the value of the `NROWS` field '
                  r'(:math:`NPPBV \cdot NBPC > NROWS`), the rows indexed with the value of the `NROWS` field '
                  r'to (:math:`NPPBV\cdot NBPC - 1`) shall contain fill data. NOTE: Only the rows indexed '
                  '0 to the value of the `NROWS` field minus 1 of the image contain significant data. '
                  'The pixel fill values are determined by the application.')  # type: int
    NCOLS = _IntegerDescriptor(
        'NCOLS', True, 8, default_value=0,
        docstring='Number of Significant Columns in Image. This field shall contain the total number of '
                  'columns of significant pixels in the image. When the product of the values of the `NPPBH` '
                  'field and the `NBPR` field is greater than the `NCOLS` field '
                  r'(:math:`NPPBH\cdot NBPR > NCOLS`), the columns indexed with the value of the `NCOLS` field '
                  r'to (:math:`NPPBH\cdot NBPR - 1`) shall contain fill data. NOTE: Only the columns '
                  'indexed 0 to the value of the `NCOLS` field minus 1 of the image contain significant data. '
                  'The pixel fill values are determined by the application.')  # type: int
    PVTYPE = _StringEnumDescriptor(
        'PVTYPE', True, 3, {'INT', 'B', 'SI', 'R', 'C'},
        docstring='Pixel Value Type. This field shall contain an indicator of the type of computer representation '
                  'used for the value for each pixel for each band in the image. ')  # type: str
    IREP = _StringEnumDescriptor(
        'IREP', True, 8,
        {'MONO', 'RGB', 'RGB/LUT', 'MULTI', 'NODISPLY', 'NVECTOR', 'POLAR', 'VPH', 'YCbCr601'},
        default_value='NODISPLY',
        docstring='Image Representation. This field shall contain a valid indicator of the processing required '
                  'in order to display an image.')  # type: str
    ICAT = _StringDescriptor(
        'ICAT', True, 8, default_value='SAR',
        docstring='Image Category. This field shall contain a valid indicator of the specific category of image, '
                  'raster or grid data. The specific category of an IS reveals its intended use or the nature '
                  'of its collector.')  # type: str
    ABPP = _IntegerDescriptor(
        'ABPP', True, 2,
        docstring='Actual Bits-Per-Pixel Per Band. This field shall contain the number of "significant bits" for '
                  'the value in each band of each pixel without compression. Even when the image is compressed, '
                  '`ABPP` contains the number of significant bits per pixel that were present in the image '
                  'before compression. This field shall be less than or equal to Number of Bits Per Pixel '
                  '(field `NBPP`). The number of adjacent bits within each `NBPP` is '
                  'used to represent the value.')  # type: int
    PJUST = _StringEnumDescriptor(
        'PJUST', True, 1, {'L', 'R'}, default_value='R',
        docstring='Pixel Justification. When `ABPP` is not equal to `NBPP`, this field indicates whether the '
                  'significant bits are left justified (:code:`L`) or right '
                  'justified (:code:`R`).')  # type: str
    ICORDS = _StringEnumDescriptor(
        'ICORDS', True, 1, {'U', 'G', 'C', 'N'}, default_value='G',
        docstring='Image Coordinate Representation. This field shall contain a valid code indicating the type '
                  'of coordinate representation used for providing an approximate location of the image in the '
                  'Image Geographic Location field (`IGEOLO`).')  # type: str
    Comments = _NITFElementDescriptor(
        'Comments', True, ImageComments, default_args={},
        docstring='The image comments.')  # type: ImageComments
    Bands = _NITFElementDescriptor(
        'Bands', True, ImageBands, default_args={},
        docstring='The image bands.')  # type: ImageBands
    ISYNC = _IntegerDescriptor(
        'ISYNC', True, 1, default_value=0,
        docstring='Image Sync code. This field is reserved for future use. ')  # type: int
    IMODE = _StringEnumDescriptor(
        'IMODE', True, 1, {'B', 'P', 'R', 'S'}, default_value='P',
        docstring='Image Mode. This field shall indicate how the Image Pixels are '
                  'stored in the NITF file.')  # type: str
    NBPR = _IntegerDescriptor(
        'NBPR', True, 4, default_value=1,
        docstring='Number of Blocks Per Row. This field shall contain the number of image blocks in a row of '
                  'blocks (paragraph 5.4.2.2) in the horizontal direction. If the image consists of only a '
                  'single block, this field shall contain the value one.')  # type: int
    NBPC = _IntegerDescriptor(
        'NBPC', True, 4, default_value=1,
        docstring='Number of Blocks Per Column. This field shall contain the number of image blocks in a column '
                  'of blocks (paragraph 5.4.2.2) in the vertical direction. If the image consists of only a '
                  'single block, this field shall contain the value one.')  # type: int
    NPPBH = _IntegerDescriptor(
        'NPPBH', True, 4, default_value=0,
        docstring='Number of Pixels Per Block Horizontal. This field shall contain the number of pixels horizontally '
                  'in each block of the image. It shall be the case that the product of the values of the `NBPR` '
                  'field and the `NPPBH` field is greater than or equal to the value of the `NCOLS` field '
                  r'(:math:`NBPR\cdot NPPBH \geq NCOLS`). When NBPR is :code:`1`, setting the `NPPBH` '
                  'value to :code:`0` designates that the number of pixels horizontally is specified by the '
                  'value in NCOLS.')  # type: int
    NPPBV = _IntegerDescriptor(
        'NPPBV', True, 4, default_value=0,
        docstring='Number of Pixels Per Block Vertical. This field shall contain the number of pixels vertically '
                  'in each block of the image. It shall be the case that the product of the values of the `NBPC` '
                  'field and the `NPPBV` field is greater than or equal to the value of the `NROWS` field '
                  r'(:math:`NBPC\cdot NPPBV \geq NROWS`). When `NBPC` is :code:`1`, setting the `NPPBV` value '
                  r'to :code:`0` designates that the number of pixels vertically is specified by '
                  r'the value in `NROWS`.')  # type: int
    NBPP = _IntegerDescriptor(
        'NBPP', True, 2, default_value=0,
        docstring='Number of Bits Per Pixel Per Band.')  # type: int
    IDLVL = _IntegerDescriptor(
        'IDLVL', True, 3, default_value=0,
        docstring='Image Display Level. This field shall contain a valid value that indicates the display level of '
                  'the image relative to other displayed file components in a composite display. The valid values '
                  'are :code:`1-999`. The display level of each displayable segment (image or graphic) within a file '
                  'shall be unique.')  # type: int
    IALVL = _IntegerDescriptor(
        'IALVL', True, 3, default_value=0,
        docstring='Attachment Level. This field shall contain a valid value that indicates the attachment '
                  'level of the image.')  # type: int
    ILOC = _StringDescriptor(
        'ILOC', True, 10, default_value='',
        docstring='Image Location. The image location is the location of the first pixel of the first line of the '
                  'image. This field shall contain the image location offset from the `ILOC` or `SLOC` value '
                  'of the segment to which the image is attached or from the origin of the CCS when the image '
                  'is unattached (`IALVL` contains :code:`0`). A row or column value of :code:`0` indicates no offset. '
                  'Positive row and column values indicate offsets down and to the right while negative row and '
                  'column values indicate offsets up and to the left.')  # type: str
    IMAG = _StringDescriptor(
        'IMAG', True, 4, default_value='1.0',
        docstring='Image Magnification. This field shall contain the magnification (or reduction) factor of the '
                  'image relative to the original source image. Decimal values are used to indicate magnification, '
                  'and decimal fraction values indicate reduction. For example, :code:`2.30` indicates the original '
                  'image has been magnified by a factor of :code:`2.30`, while :code:`0.5` indicates '
                  'the original image has been reduced by a factor of 2.')  # type: str
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType
    ExtendedHeader = _NITFElementDescriptor(
        'ExtendedHeader', True, UserHeaderType, default_args={},
        docstring='Extended subheader - TRE list.')  # type: UserHeaderType

    def __init__(self, **kwargs):
        self._IC = None
        self._COMRAT = None
        self._IGEOLO = None
        self._mask_subheader = None
        super(ImageSegmentHeader0, self).__init__(**kwargs)

    @property
    def is_masked(self):
        """
        bool: Does this image segment contain a mask?
        """

        return self.IC in ['NM', 'M1', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8']

    @property
    def is_compressed(self):
        """
        bool: Is this image segment compressed?
        """

        return self.IC not in ['NC', 'NM']

    @property
    def IC(self):
        """
        str: Image Compression. This field shall contain a valid code indicating
        the form of compression used in representing the image data.

        Valid values for this field are, :code:`C1` to represent bi-level, :code:`C3`
        to represent JPEG, :code:`C4` to represent Vector Quantization, :code:`C5`
        to represent lossless JPEG, :code:`I1` to represent down sampled JPEG,
        and :code:`NC` to represent the image is not compressed. Also valid are
        :code:`M1, M3, M4`, and :code:`M5` for compressed images, and :code:`NM`
        for uncompressed images indicating an image that contains a block
        mask and/or a pad pixel mask. :code:`C6` and :code:`M6` are reserved values
        that will represent a future correlated multicomponent compression
        algorithm. :code:`C7` and :code:`M7` are reserved values that will represent
        a future complex SAR compression. :code:`C8` and :code:`M8` are the values
        for ISO standard compression JPEG 2000.

        The format of a mask image is identical to the format of its corresponding non-masked image
        except for the presence of an Image Data Mask at the beginning of
        the image data area. The format of the Image Data Mask is described
        in paragraph 5.4.3.2 and is shown in table A-3(A). The definitions
        of the compression schemes associated with codes :code:`C1/M1, C3/M3, C4/M4, C5/M5`
        are given, respectively, in ITU- T T.4, AMD2, MIL-STD-188-198A,
        MIL-STD- 188-199, and NGA N0106-97. :code:`C1` is found in ITU- T T.4 AMD2,
        :code:`C3` is found in MIL-STD-188-198A, :code:`C4` is found in MIL-STD-188-199,
        and :code:`C5` and :code:`I1` are found in NGA N0106-97. (NOTE: :code:`C2` (ARIDPCM) is not
        valid in NITF 2.1.) The definition of the compression scheme associated
        with codes :code:`C8/M8` is found in ISO/IEC 15444- 1:2000 (with amendments 1 and 2).
        """

        return self._IC

    @IC.setter
    def IC(self, value):
        value = _parse_str(value, 2, 'NC', 'IC', self)
        if value not in {
                'NC', 'NM', 'C1', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'I1',
                'M1', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8'}:
            raise ValueError('IC got invalid value {}'.format(value))
        self._IC = value
        if value in ('NC', 'NM'):
            self._COMRAT = None
        elif self._COMRAT is not None:
            self._COMRAT = '\x20'*4

    @property
    def COMRAT(self):
        """
        None|str: Compression Rate Code. If the IC field contains one of
        :code:`C1, C3, C4, C5, C8, M1, M3, M4, M5, M8, I1`, this field shall be contain
        a code indicating the compression rate for the image.

        If `IC` is :code:`NC` or :code:`NM`, then this will be set to :code:`None`.
        """

        return self._COMRAT

    @COMRAT.setter
    def COMRAT(self, value):
        value = _parse_str(value, 4, None, 'COMRAT', self)
        if value is None and self.IC not in ('NC', 'NM'):
            value = '\x20'*4
            logger.error(
                'COMRAT value is None, but IC is not in {"NC", "NM"}.\n\t'
                'This must be resolved.')
        if value is not None and self.IC in ('NC', 'NM'):
            value = None
            logger.error(
                'COMRAT value is something other than None, but IC in {"NC", "NM"}.\n\t'
                'This is invalid, and COMRAT is being set to None.')
        self._COMRAT = value

    @property
    def IGEOLO(self):
        """
        None|str: Image Geographic Location. This field, when present, shall contain
        an approximate geographic location which is not intended for analytical purposes
        (e.g., targeting, mensuration, distance calculation); it is intended to support
        general user appreciation for the image location (e.g., cataloguing). The
        representation of the image corner locations is specified in the `ICORDS` field.
        The locations of the four corners of the (significant) image data shall be given
        in image coordinate order: (0,0), (0, MaxCol), (MaxRow, MaxCol), (MaxRow, 0).
        MaxCol and MaxRow shall be determined from the values contained, respectively,
        in the `NCOLS` field and the `NROWS` field.
        """

        return self._IGEOLO

    @IGEOLO.setter
    def IGEOLO(self, value):
        value = _parse_str(value, 60, None, 'IGEOLO', self)
        if value is None and self.ICORDS.strip() != '':
            value = '\x20'*60
        if value is not None and self.ICORDS.strip() == '':
            value = None
        self._IGEOLO = value

    @property
    def mask_subheader(self):
        # type: () -> Union[None, MaskSubheader]
        """
        None|MaskSubheader: The mask subheader, if it has been appended.
        """

        return self._mask_subheader

    @mask_subheader.setter
    def mask_subheader(self, value):
        if value is None:
            self._mask_subheader = None
            return
        if not isinstance(value, MaskSubheader):
            raise ValueError(
                'mask_subheader is expected to be an instance of MaskSubheader. '
                'Got type {}'.format(type(value)))
        if self.IC not in ['NM', 'M1', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8']:
            raise ValueError(
                'IC={}, which does not indicate the presence of a mask '
                'subheader'.format(self.IC))
        self._mask_subheader = value

    def _get_attribute_length(self, fld):
        if fld in ['COMRAT', 'IGEOLO']:
            if getattr(self, '_'+fld) is None:
                return 0
            else:
                return self._lengths[fld]
        else:
            return super(ImageSegmentHeader0, self)._get_attribute_length(fld)

    @classmethod
    def minimum_length(cls):
        # COMRAT and IGEOLO may not be there
        return super(ImageSegmentHeader0, cls).minimum_length() - 64

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'IC':
            val = value[start:start+2].decode('utf-8')
            fields['IC'] = val
            if val in ('NC', 'NM'):
                fields['COMRAT'] = None
            out = start+2
        elif attribute == 'ICORDS':
            fields['ICORDS'] = value[start:start+1]
            if fields['ICORDS'] == b'N':
                fields['IGEOLO'] = None
            out = start+1
        else:
            out = super(ImageSegmentHeader0, cls)._parse_attribute(fields, attribute, value, start)
        return out

    def get_uncompressed_block_size(self) -> int:
        """
        Gets the size of an uncompressed block.

        Note that if `IMODE == 'S'`, then each block consists of a single band.
        Otherwise, a block consists of all bands.

        Returns
        -------
        int
        """

        nppbv = self.NROWS if self.NPPBV == 0 else self.NPPBV
        nppbh = self.NCOLS if self.NPPBH == 0 else self.NPPBH
        if self.IMODE == 'S':
            return int(nppbh*nppbv*self.NBPP/8)
        else:
            return int(nppbh*nppbv*len(self.Bands)*self.NBPP/8)

    def get_full_uncompressed_image_size(self) -> int:
        """
        Gets the full size in bytes of the uncompressed image including any padding
        in the blocks.

        Returns
        -------
        int
        """

        total_blocks = self.NBPR*self.NBPC
        if self.IMODE == 'S':
            total_blocks *= len(self.Bands)
        return total_blocks*self.get_uncompressed_block_size()

    def get_clevel(self) -> int:
        """
        Gets the CLEVEL value for this image segment.

        Returns
        -------
        int
        """

        dim_size = max(self.NROWS, self.NCOLS)
        if dim_size <= 2048:
            return 3
        elif dim_size <= 8192:
            return 5
        elif dim_size <= 65536:
            return 6
        else:
            return 7
